"""
Support for functionality to have conversations with AI-Speaker.

"""
import asyncio
import logging
import re
import warnings
import json
import voluptuous as vol
import datetime
import requests
import os
from homeassistant import core
from homeassistant.loader import bind_hass
from homeassistant.const import (
    ATTR_ENTITY_ID, SERVICE_TURN_OFF,
    SERVICE_TURN_ON, ATTR_UNIT_OF_MEASUREMENT, SERVICE_OPEN_COVER, SERVICE_CLOSE_COVER,
    STATE_ON, STATE_OFF, STATE_HOME, STATE_NOT_HOME, STATE_UNKNOWN, STATE_OPEN, STATE_OPENING, STATE_CLOSED,
    STATE_CLOSING, STATE_PLAYING, STATE_PAUSED, STATE_IDLE, STATE_STANDBY, STATE_ALARM_DISARMED, STATE_ALARM_ARMED_HOME,
    STATE_ALARM_ARMED_AWAY, STATE_ALARM_ARMED_NIGHT, STATE_ALARM_ARMED_CUSTOM_BYPASS, STATE_ALARM_PENDING,
    STATE_ALARM_ARMING, STATE_ALARM_DISARMING, STATE_ALARM_TRIGGERED, STATE_LOCKED, STATE_UNLOCKED,
    STATE_UNAVAILABLE, STATE_OK, STATE_PROBLEM, ATTR_ASSUMED_STATE)
from homeassistant.helpers import intent, config_validation as cv
from homeassistant.components import ais_cloud
import homeassistant.components.mqtt as mqtt
from homeassistant.components import ais_drives_service
import homeassistant.ais_dom.ais_global as ais_global
aisCloudWS = ais_cloud.AisCloudWS()

REQUIREMENTS = ['fuzzywuzzy==0.15.1', 'babel']
# DEPENDENCIES = ['http']

ATTR_TEXT = 'text'
DOMAIN = 'ais_ai_service'
G_HTTP_REST_SERVICE_BASE_URL = 'http://{}:8122'

REGEX_TURN_COMMAND = re.compile(r'turn (?P<name>(?: |\w)+) (?P<command>\w+)')

SERVICE_PROCESS_SCHEMA = vol.Schema({
    vol.Required(ATTR_TEXT): cv.string,
})

CONFIG_SCHEMA = vol.Schema({DOMAIN: vol.Schema({
    vol.Optional('intents'): vol.Schema({
        cv.string: vol.All(cv.ensure_list, [cv.string])
    })
})}, extra=vol.ALLOW_EXTRA)

INTENT_GET_TIME = 'AisGetTime'
INTENT_GET_DATE = 'AisGetDate'
INTENT_PLAY_RADIO = 'AisPlayRadio'
INTENT_PLAY_PODCAST = 'AisPlayPodcast'
INTENT_PLAY_YT_MUSIC = 'AisPlayYtMusic'
INTENT_ASK_QUESTION = 'AisAskQuestion'
INTENT_CHANGE_CONTEXT = 'AisChangeContext'
INTENT_GET_WEATHER = 'AisGetWeather'
INTENT_GET_WEATHER_48 = 'AisGetWeather48'
INTENT_STATUS = 'AisStatusInfo'
INTENT_TURN_ON = 'AisTurnOn'
INTENT_TURN_OFF = 'AisTurnOff'
INTENT_LAMPS_ON = 'AisLampsOn'
INTENT_LAMPS_OFF = 'AisLampsOff'
INTENT_SWITCHES_ON = 'AisSwitchesOn'
INTENT_SWITCHES_OFF = 'AisSwitchesOff'
INTENT_OPEN_COVER = 'AisCoverOpen'
INTENT_CLOSE_COVER = 'AisCoverClose'
INTENT_STOP = 'AisStop'
INTENT_PLAY = 'AisPlay'
INTENT_NEXT = 'AisNext'
INTENT_PREV = 'AisPrev'
INTENT_SCENE = 'AisSceneActive'
INTENT_SAY_IT = 'AisSayIt'
INTENT_CLIMATE_SET_TEMPERATURE = 'AisClimateSetTemperature'
INTENT_CLIMATE_SET_AWAY = 'AisClimateSetAway'
INTENT_CLIMATE_UNSET_AWAY = 'AisClimateUnSetAway'
INTENT_CLIMATE_SET_ALL_ON = 'AisClimateSetAllOn'
INTENT_CLIMATE_SET_ALL_OFF = 'AisClimateSetAllOff'

REGEX_TYPE = type(re.compile(''))

_LOGGER = logging.getLogger(__name__)
GROUP_VIEWS = ['Pomoc', 'Twój Dom', 'Audio', 'Ustawienia']
CURR_GROUP_VIEW = None
# group entities in each group view, see main_ais_groups.yaml
GROUP_ENTITIES = []
ALL_AIS_SENSORS = []
CURR_GROUP = None
CURR_ENTITIE = None
CURR_ENTITIE_ENTERED = False
CURR_BUTTON_CODE = None
CURR_BUTTON_LONG_PRESS = False
CURR_ENTITIE_POSITION = None
ALL_SWITCHES = ["input_boolean", "automation", "switch", "light", "media_player", "script"]

# ais-dom virtual keyboard
# kodowała to Asia Raczkowska w 2019 roku
VIRTUAL_KEYBOARD_MODE = ['Litery', 'Wielkie litery', 'Cyfry', 'Znaki specjalne', 'Usuwanie']
CURR_VIRTUAL_KEYBOARD_MODE = None
VIRTUAL_KEYBOARD_LETTERS = ['-', 'A', 'Ą', 'B', 'C', 'Ć', 'D', 'E', 'Ę', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'Ł', 'M',
                            'N', 'Ń', 'O', 'Ó', 'P', 'Q', 'R', 'S', 'Ś', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'Ź', 'Ż']
VIRTUAL_KEYBOARD_NUMBERS = ['-', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
VIRTUAL_KEYBOARD_SYMBOLS = ['-', ' ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '_', '.', '/',
                            ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '{', '|', '}']
VIRTUAL_KEYBOARD_SYMBOLS_NAMES = ['-', 'spacja', 'wykrzyknik', 'cudzysłów', 'hash', 'dolar', 'procent', 'symbol and',
                                  'pojedynczy cudzysłów', 'nawias otwierający', 'nawias zamykający', 'gwiazdka',
                                  'plus', 'przecinek', 'myślnik', 'podkreślenie dolne', 'kropka', 'ukośnik prawy',
                                  'dwukropek', 'średnik', 'znak mniejszości', 'znak równości', 'znak większości',
                                  'znak zapytania', 'małpa', 'kwadratowy nawias otwierający', 'ukośnik lewy',
                                  'kwadratowy nawias zamykający', 'daszek', 'nawias klamrowy otwierający',
                                  'kreska pionowa', 'nawias klamrowy zamykający']
VIRTUAL_KEYBOARD_DELETE = ['-', 'ostatni znak', 'ostatni wyraz', 'całe pole']
CURR_VIRTUAL_KEYBOARD_VALUE = None
CURR_VIRTUAL_KEY = None
# ais-dom virtual keyboard

GLOBAL_TTS_TEXT = None

def get_tts_text():
    return GLOBAL_TTS_TEXT


def isSwitch(entity_id):
    global ALL_SWITCHES
    # problem with startswith and tuple of strings
    for s in ALL_SWITCHES:
        if entity_id.startswith(s):
            return True
    return False


@core.callback
@bind_hass
def async_register(hass, intent_type, utterances):
    """Register an intent.
    Registrations don't require conversations to be loaded. They will become
    active once the conversation component is loaded.
    """
    intents = hass.data.get(DOMAIN)

    if intents is None:
        intents = hass.data[DOMAIN] = {}

    conf = intents.get(intent_type)

    if conf is None:
        conf = intents[intent_type] = []

    for utterance in utterances:
        if isinstance(utterance, REGEX_TYPE):
            conf.append(utterance)
        else:
            conf.append(_create_matcher(utterance))


def translate_state(info_data):
    if not info_data:
        info_data = ""
    elif info_data == STATE_ON:
        info_data = "włączone"
    elif info_data == STATE_OFF:
        info_data = "wyłączone"
    elif info_data == STATE_HOME:
        info_data = "w domu"
    elif info_data == STATE_NOT_HOME:
        info_data = "poza domem"
    elif info_data == STATE_UNKNOWN:
        info_data = "status nieznany"
    elif info_data == STATE_OPEN:
        info_data = "otwarty"
    elif info_data == STATE_OPENING:
        info_data = "otwieranie"
    elif info_data == STATE_CLOSED:
        info_data = "zamknięty"
    elif info_data == STATE_CLOSING:
        info_data = "zamykanie"
    elif info_data == STATE_PAUSED:
        info_data = "pauza"
    elif info_data == STATE_PLAYING:
        info_data = "odtwarzanie"
    elif info_data == STATE_IDLE:
        info_data = "status bezczynny"
    elif info_data == STATE_STANDBY:
        info_data = "status bezczynny"
    elif info_data == STATE_ALARM_DISARMED:
        info_data = "status rozbrojony"
    elif info_data == STATE_ALARM_ARMED_HOME:
        info_data = "status uzbrojony w domu"
    elif info_data == STATE_ALARM_ARMED_AWAY:
        info_data = "status uzbrojony poza domem"
    elif info_data == STATE_ALARM_ARMED_NIGHT:
        info_data = "status uzbrojony noc"
    elif info_data == STATE_ALARM_ARMED_CUSTOM_BYPASS:
        info_data = "status uzbrojony własny"
    elif info_data == STATE_ALARM_ARMING:
        info_data = "alarm uzbrajanie"
    elif info_data == STATE_ALARM_DISARMING:
        info_data = "alarm rozbrajanie"
    elif info_data == STATE_ALARM_TRIGGERED:
        info_data = "alarm powiadomiony"
    elif info_data == STATE_LOCKED:
        info_data = "zamknięty"
    elif info_data == STATE_UNLOCKED:
        info_data = "otwarty"
    elif info_data == STATE_UNAVAILABLE:
        info_data = "niedostępny"
    elif info_data == STATE_OK:
        info_data = "ok"
    elif info_data == STATE_PROBLEM:
        info_data = "problem"
    elif info_data == "above_horizon":
        info_data = "powyżej horyzontu"
    elif info_data == "below_horizon":
        info_data = "poniżej horyzontu"

    return info_data


def get_next(arra, curr):
    _first = None
    _curr = None
    _next = None
    for a in arra:
        # ignore empy option
        if a != ais_global.G_EMPTY_OPTION:
            if _curr is not None and _next is None:
                _next = a
            if _first is None:
                _first = a
            if curr == a:
                _curr = a
    if _next is not None:
        return _next
    else:
        return _first


def get_prev(arra, curr):
    _last = None
    _curr = None
    _prev = None
    for a in arra:
        # ignore empy option
        if a != ais_global.G_EMPTY_OPTION:
            _last = a
            if curr == a:
                _curr = a
            if _curr is None:
                _prev = a
    if _prev is not None:
        return _prev
    else:
        return _last


# Group views: Dom -> Audio -> Ustawienia -> Pomoc
def get_curr_group_view():
    if CURR_GROUP_VIEW is None:
        return GROUP_VIEWS[0]
    return CURR_GROUP_VIEW


def say_curr_group_view(hass):
    _say_it(hass, get_curr_group_view(), None)


def set_curr_group_view():
    # set focus on current menu group view
    global CURR_GROUP_VIEW
    global CURR_GROUP
    global CURR_ENTITIE
    global CURR_ENTITIE_ENTERED
    global CURR_ENTITIE_POSITION
    CURR_GROUP = None
    CURR_ENTITIE = None
    CURR_ENTITIE_ENTERED = False
    CURR_ENTITIE_POSITION = None
    CURR_GROUP_VIEW = get_curr_group_view()


def set_next_group_view():
    # set focus on next menu group view
    global CURR_GROUP_VIEW
    CURR_GROUP_VIEW = get_next(GROUP_VIEWS, get_curr_group_view())
    # to reset
    set_curr_group_view()


def set_prev_group_view():
    # set focus on prev menu group view
    global CURR_GROUP_VIEW
    CURR_GROUP_VIEW = get_prev(GROUP_VIEWS, get_curr_group_view())
    # to reset
    set_curr_group_view()


# virtual keybord
# Group views: Litery -> Wielkie litery -> Cyfry -> Znaki specjalne -> Usuwanie
def get_curr_virtual_keyboard_mode():
    if CURR_VIRTUAL_KEYBOARD_MODE is None:
        return VIRTUAL_KEYBOARD_MODE[0]
    return CURR_VIRTUAL_KEYBOARD_MODE


def set_next_virtual_keyboard_mode():
    global CURR_VIRTUAL_KEYBOARD_MODE
    global CURR_VIRTUAL_KEY
    CURR_VIRTUAL_KEY = None
    CURR_VIRTUAL_KEYBOARD_MODE = get_next(VIRTUAL_KEYBOARD_MODE, get_curr_virtual_keyboard_mode())


def set_prev_virtual_keyboard_mode():
    global CURR_VIRTUAL_KEYBOARD_MODE
    global CURR_VIRTUAL_KEY
    CURR_VIRTUAL_KEY = None
    CURR_VIRTUAL_KEYBOARD_MODE = get_prev(VIRTUAL_KEYBOARD_MODE, get_curr_virtual_keyboard_mode())


def say_curr_virtual_keyboard_mode(hass):
    _say_it(hass, get_curr_virtual_keyboard_mode(), None)


def get_curr_virtual_key():
    if CURR_VIRTUAL_KEY is not None:
        return str(CURR_VIRTUAL_KEY)
    km = get_curr_virtual_keyboard_mode()
    if km == "Litery":
        return VIRTUAL_KEYBOARD_LETTERS[0]
    elif km == "Wielkie litery":
        return VIRTUAL_KEYBOARD_LETTERS[0]
    elif km == "Cyfry":
        return VIRTUAL_KEYBOARD_NUMBERS[0]
    elif km == "Znaki specjalne":
        return VIRTUAL_KEYBOARD_SYMBOLS[0]
    elif km == "Usuwanie":
        return VIRTUAL_KEYBOARD_DELETE[0]


def set_next_virtual_key():
    global CURR_VIRTUAL_KEY
    km = get_curr_virtual_keyboard_mode()
    if km == "Litery":
        CURR_VIRTUAL_KEY = get_next(VIRTUAL_KEYBOARD_LETTERS, get_curr_virtual_key())
    elif km == "Wielkie litery":
        CURR_VIRTUAL_KEY = get_next(VIRTUAL_KEYBOARD_LETTERS, get_curr_virtual_key())
    elif km == "Cyfry":
        CURR_VIRTUAL_KEY = get_next(VIRTUAL_KEYBOARD_NUMBERS, get_curr_virtual_key())
    elif km == "Znaki specjalne":
        CURR_VIRTUAL_KEY = get_next(VIRTUAL_KEYBOARD_SYMBOLS, get_curr_virtual_key())
    elif km == "Usuwanie":
        CURR_VIRTUAL_KEY = get_next(VIRTUAL_KEYBOARD_DELETE, get_curr_virtual_key())


def set_prev_virtual_key():
    global CURR_VIRTUAL_KEY
    km = get_curr_virtual_keyboard_mode()
    if km == "Litery":
        CURR_VIRTUAL_KEY = get_prev(VIRTUAL_KEYBOARD_LETTERS, get_curr_virtual_key())
    elif km == "Wielkie litery":
        CURR_VIRTUAL_KEY = get_prev(VIRTUAL_KEYBOARD_LETTERS, get_curr_virtual_key())
    elif km == "Cyfry":
        CURR_VIRTUAL_KEY = get_prev(VIRTUAL_KEYBOARD_NUMBERS, get_curr_virtual_key())
    elif km == "Znaki specjalne":
        CURR_VIRTUAL_KEY = get_prev(VIRTUAL_KEYBOARD_SYMBOLS, get_curr_virtual_key())
    elif km == "Usuwanie":
        CURR_VIRTUAL_KEY = get_prev(VIRTUAL_KEYBOARD_DELETE, get_curr_virtual_key())


def say_curr_virtual_key(hass):
    key = get_curr_virtual_key()
    km = get_curr_virtual_keyboard_mode()
    text = ""
    if km == "Litery":
        text = "" + key.lower()
    elif km == "Wielkie litery":
        text = "" + key
    elif km == "Cyfry":
        text = "" + key
    elif km == "Znaki specjalne":
        idx = VIRTUAL_KEYBOARD_SYMBOLS.index(key)
        text = "" + VIRTUAL_KEYBOARD_SYMBOLS_NAMES[idx]
    elif km == "Usuwanie":
        text = "" + key

    _say_it(hass, text, None)


def reset_virtual_keyboard(hass):
    global CURR_VIRTUAL_KEYBOARD_MODE
    global CURR_VIRTUAL_KEY
    global CURR_VIRTUAL_KEYBOARD_VALUE
    CURR_VIRTUAL_KEYBOARD_MODE = None
    CURR_VIRTUAL_KEY = None
    CURR_VIRTUAL_KEYBOARD_VALUE = None
    # reset field value
    hass.services.call('input_text', 'set_value', {
        "entity_id": CURR_ENTITIE, "value": ""})



# Groups in Groups views
def get_curr_group():
    global CURR_GROUP
    if CURR_GROUP is None:
        # take the first one from Group view
        for group in GROUP_ENTITIES:
            if group['remote_group_view'] == get_curr_group_view():
                CURR_GROUP = group
    return CURR_GROUP


def get_curr_group_idx():
    idx = 0
    for group in GROUP_ENTITIES:
        if group['entity_id'] == get_curr_group()['entity_id']:
            return idx
        idx += 1
    return idx


def say_curr_group(hass):
    _say_it(hass, get_curr_group()['friendly_name'], None)


def set_bookmarks_curr_group(hass):
    for idx, g in enumerate(GROUP_ENTITIES, start=0):
        if g['entity_id'] == 'group.bookmarks':
            set_curr_group(hass, g)


def set_curr_group(hass, group):
    # set focus on current menu group view
    global CURR_GROUP_VIEW
    global CURR_GROUP
    global CURR_ENTITIE
    global CURR_ENTITIE_ENTERED
    global CURR_ENTITIE_POSITION
    # the entitie can be selected or focused
    CURR_ENTITIE = None
    CURR_ENTITIE_ENTERED = False
    CURR_ENTITIE_POSITION = None
    if group is None:
        CURR_GROUP = get_curr_group()

        hass.states.async_set(
            'binary_sensor.selected_entity',
            CURR_GROUP['entity_id'])
    else:
        CURR_GROUP_VIEW = group['remote_group_view']
        CURR_GROUP = group


def set_next_group(hass):
    # set focus on next group in focused view
    global CURR_GROUP
    first_group_in_view = None
    curr_group_in_view = None
    next_group_in_view = None
    for group in GROUP_ENTITIES:
        if group['remote_group_view'] == get_curr_group_view():
            # select the first group
            if curr_group_in_view is not None and next_group_in_view is None:
                next_group_in_view = group
            if first_group_in_view is None:
                first_group_in_view = group
            if CURR_GROUP['entity_id'] == group['entity_id']:
                curr_group_in_view = group

    if next_group_in_view is not None:
        CURR_GROUP = next_group_in_view
    else:
        CURR_GROUP = first_group_in_view
    # to reset
    set_curr_group(hass, CURR_GROUP)


def set_prev_group(hass):
    # set focus on prev group in focused view
    global CURR_GROUP
    last_group_in_view = None
    curr_group_in_view = None
    prev_group_in_view = None
    for group in GROUP_ENTITIES:
        if group['remote_group_view'] == get_curr_group_view():
            # select the last group
            last_group_in_view = group
            if CURR_GROUP['entity_id'] == group['entity_id']:
                curr_group_in_view = group
            if curr_group_in_view is None:
                prev_group_in_view = group
    if prev_group_in_view is not None:
        CURR_GROUP = prev_group_in_view
    else:
        CURR_GROUP = last_group_in_view
    # to reset
    set_curr_group(hass, CURR_GROUP)


# entity in group
def get_curr_entity():
    global CURR_ENTITIE
    if CURR_ENTITIE is None:
        CURR_ENTITIE = GROUP_ENTITIES[get_curr_group_idx()]['entities'][0]
    return CURR_ENTITIE


def get_curr_entity_idx():
    idx = 0
    for item in GROUP_ENTITIES[get_curr_group_idx()]['entities']:
        if item == get_curr_entity():
            return idx
        idx += 1


def set_curr_entity(hass, entity):
    # set focus on current entity
    global CURR_ENTITIE
    global CURR_ENTITIE_POSITION
    if entity is None:
        CURR_ENTITIE = get_curr_entity()
    else:
        CURR_ENTITIE = entity
    CURR_ENTITIE_POSITION = None
    hass.states.async_set(
        'binary_sensor.selected_entity',
        CURR_ENTITIE)


def set_next_entity(hass):
    # set next entity
    global CURR_ENTITIE
    entity_idx = get_curr_entity_idx()
    group_idx = get_curr_group_idx()
    l_group_len = len(GROUP_ENTITIES[group_idx]['entities'])
    if entity_idx + 1 == l_group_len:
        entity_idx = 0
    else:
        entity_idx = entity_idx + 1
    CURR_ENTITIE = GROUP_ENTITIES[group_idx]['entities'][entity_idx]
    # to reset variables
    set_curr_entity(hass, None)
    say_curr_entity(hass)


def set_prev_entity(hass):
    # set prev entity
    global CURR_ENTITIE
    idx = get_curr_entity_idx()
    l_group_len = len(GROUP_ENTITIES[get_curr_group_idx()]['entities'])
    if idx == 0:
        idx = l_group_len - 1
    else:
        idx = idx - 1
    CURR_ENTITIE = GROUP_ENTITIES[get_curr_group_idx()]['entities'][idx]
    # to reset variables
    set_curr_entity(hass, None)
    say_curr_entity(hass)


def say_curr_entity(hass):
    # check if we have selected item
    entity_id = get_curr_entity()
    state = hass.states.get(entity_id)
    if state is None:
        _say_it(hass, "Brak pozycji", None)
        return
    text = state.attributes.get('text')
    info_name = state.attributes.get('friendly_name')
    info_data = state.state
    info_unit = state.attributes.get('unit_of_measurement')
    if not text:
        text = ""
    # handle special cases...
    if entity_id == "sensor.ais_knowledge_answer":
        _say_it(hass, "Odpowiedź: " + text, None)
        return
    elif entity_id == 'input_select.ais_bookmark_last_played':
        _say_it(hass, info_name + " " + info_data.replace("Local;", ""), None)
        return
    elif entity_id.startswith('script.'):
        _say_it(hass, info_name + " Naciśnij OK/WYKONAJ by uruchomić.", None)
        return
    elif entity_id.startswith('input_text.'):
        if CURR_BUTTON_CODE == 4:
            if CURR_VIRTUAL_KEYBOARD_VALUE is None:
                _say_it(hass, "Nic nie wpisałeś", None)
            else:
                _say_it(hass, "Wpisałeś " + CURR_VIRTUAL_KEYBOARD_VALUE, None)
        else:
            _say_it(hass, info_name + " " + info_data + ". Naciśnij OK by wpisać.", None)
        return
    elif entity_id.startswith('input_select.'):
        if CURR_BUTTON_CODE == 4:
            if info_data == ais_global.G_EMPTY_OPTION:
                _say_it(hass, "Brak wyboru", None)
            else:
                _say_it(hass, "Wybrałeś " + info_data, None)
        else:
            if info_data != ais_global.G_EMPTY_OPTION:
                _say_it(hass, info_name + " " + info_data + ". Naciśnij OK by zmienić.", None)
            else:
                _say_it(hass, info_name + " " + info_data + ". Naciśnij OK by wybrać.", None)
        return
    # normal case
    # decode None
    if not info_name:
        info_name = ""
    info_data = translate_state(info_data)
    if not info_unit:
        info_unit = ""
    info = "%s %s %s" % (info_name, info_data, info_unit)
    _say_it(hass, info, None)


def get_curent_position(hass):
    # return the entity focused position
    global CURR_ENTITIE_POSITION
    if CURR_ENTITIE_POSITION is None:
        CURR_ENTITIE_POSITION = hass.states.get(CURR_ENTITIE).state
    return CURR_ENTITIE_POSITION


def commit_current_position(hass):
    if CURR_ENTITIE.startswith('input_select.'):
        # force the change - to trigger the state change for automation
        possition = get_curent_position(hass)
        hass.services.call(
            'input_select',
            'select_option', {
                "entity_id": CURR_ENTITIE,
                "option": ais_global.G_EMPTY_OPTION})
        hass.block_till_done()
        hass.services.call(
            'input_select',
            'select_option', {
                "entity_id": CURR_ENTITIE,
                "option": possition})
    elif CURR_ENTITIE.startswith('input_number.'):
        hass.services.call(
            'input_number',
            'set_value', {
                "entity_id": CURR_ENTITIE,
                "value": get_curent_position(hass)})

    if CURR_ENTITIE == "input_select.ais_android_wifi_network":
        _say_it(hass, "wybrano wifi: " + get_curent_position(hass).split(';')[0], None)
    # elif CURR_ENTITIE == "input_select.folder_name":
    #     _say_it(hass, "wybrano", None)
    else:
        _beep_it(hass, 33)
    # TODO - run the script for the item,
    # the automation on state should be executed only from app not from remote


def set_next_position(hass):
    global CURR_ENTITIE_POSITION
    CURR_ENTITIE_POSITION = get_curent_position(hass)
    state = hass.states.get(CURR_ENTITIE)
    if CURR_ENTITIE.startswith('input_select.'):
        arr = state.attributes.get('options')
        # the "-" option is always first
        if (len(arr) < 2) or (CURR_ENTITIE == "input_select.folder_name" and len(arr) < 3):
            _say_it(hass, "brak pozycji", None)
        else:
            CURR_ENTITIE_POSITION = get_next(arr, CURR_ENTITIE_POSITION)
            if CURR_ENTITIE == "input_select.folder_name":
                if CURR_ENTITIE_POSITION == "..":
                    CURR_ENTITIE_POSITION = get_next(arr, CURR_ENTITIE_POSITION)
                # check if we have remote or local path
                if CURR_ENTITIE_POSITION.startswith(ais_drives_service.G_CLOUD_PREFIX):
                    if CURR_ENTITIE_POSITION == ais_drives_service.G_CLOUD_PREFIX:
                        item = ais_drives_service.G_CLOUD_PREFIX.replace("/", "")
                    else:
                        item = CURR_ENTITIE_POSITION.replace(ais_drives_service.G_CLOUD_PREFIX, "")
                        k = item.find(":")
                        if k+1 < len(item):
                            item = item[k:]
                    _say_it(hass, os.path.basename(item), None)
                else:
                    _say_it(hass, os.path.basename(CURR_ENTITIE_POSITION), None)
            else:
                _say_it(hass, CURR_ENTITIE_POSITION, None)
    elif CURR_ENTITIE.startswith('input_number.'):
        _max = float(state.attributes.get('max'))
        _step = float(state.attributes.get('step'))
        _curr = float(CURR_ENTITIE_POSITION)
        CURR_ENTITIE_POSITION = str(round(min(_curr+_step, _max), 2))
        _say_it(hass, str(CURR_ENTITIE_POSITION), None)


def set_prev_position(hass):
    global CURR_ENTITIE_POSITION
    CURR_ENTITIE_POSITION = get_curent_position(hass)
    state = hass.states.get(CURR_ENTITIE)
    if CURR_ENTITIE.startswith('input_select.'):
        arr = state.attributes.get('options')
        if (len(arr) < 2) or (CURR_ENTITIE == "input_select.folder_name" and len(arr) < 3):
            _say_it(hass, "brak pozycji", None)
        else:
            CURR_ENTITIE_POSITION = get_prev(arr, CURR_ENTITIE_POSITION)
            if CURR_ENTITIE == "input_select.folder_name":
                if CURR_ENTITIE_POSITION == "..":
                    CURR_ENTITIE_POSITION = get_prev(arr, CURR_ENTITIE_POSITION)
                # check if we have remote or local path
                if CURR_ENTITIE_POSITION.startswith(ais_drives_service.G_CLOUD_PREFIX):
                    if CURR_ENTITIE_POSITION == ais_drives_service.G_CLOUD_PREFIX:
                        item = ais_drives_service.G_CLOUD_PREFIX.replace("/", "")
                    else:
                        item = CURR_ENTITIE_POSITION.replace(ais_drives_service.G_CLOUD_PREFIX, "")
                        k = item.find(":")
                        if k+1 < len(item):
                            item = item[k:]
                    _say_it(hass, os.path.basename(item), None)
                else:
                    _say_it(hass, os.path.basename(CURR_ENTITIE_POSITION), None)
            else:
                _say_it(hass, CURR_ENTITIE_POSITION, None)
    elif CURR_ENTITIE.startswith('input_number.'):
        _min = float(state.attributes.get('min'))
        _step = float(state.attributes.get('step'))
        _curr = float(CURR_ENTITIE_POSITION)
        CURR_ENTITIE_POSITION = str(round(max(_curr-_step, _min), 2))
        _say_it(hass, str(CURR_ENTITIE_POSITION), None)


def select_entity(hass, long_press):
    # on remote OK, select group view, group or entity
    global CURR_ENTITIE_ENTERED
    # OK on remote
    if CURR_GROUP_VIEW is None:
        # no group view was selected
        set_curr_group_view()
        say_curr_group_view(hass)
        return
    if CURR_GROUP is None:
        # no group is selected - we need to select the first one
        # from the group view
        set_curr_group(hass, None)
        say_curr_group(hass)
        return
    if CURR_ENTITIE is None:
        # no entity is selected - we need to focus the first one
        set_curr_entity(hass, None)
        say_curr_entity(hass)
        CURR_ENTITIE_ENTERED = False
        return

    if CURR_ENTITIE_ENTERED is False:
        # check if the entity option can be selected
        if can_entity_be_changed(hass, CURR_ENTITIE):
            if can_entity_be_entered(hass, CURR_ENTITIE):
                CURR_ENTITIE_ENTERED = True
                if CURR_ENTITIE.startswith('input_text.'):
                    _say_it(hass, "Wpisywanie tekstu włączone", None)
                    reset_virtual_keyboard(hass)
                else:
                    set_next_position(hass)
                return
            else:
                # we will change this item directly
                if CURR_ENTITIE.startswith("media_player."):
                    # enter to media player
                    CURR_ENTITIE_ENTERED = True
                    # play / pause on selected player
                    curr_state = hass.states.get(CURR_ENTITIE).state
                    if curr_state == 'playing':
                        if long_press is True:
                            _say_it(hass, "stop", None)
                            hass.services.call('media_player', 'media_stop', {"entity_id": CURR_ENTITIE})
                        else:
                            _say_it(hass, "pauza", None)
                            hass.services.call('media_player', 'media_pause', {"entity_id": CURR_ENTITIE})
                    else:
                        _say_it(hass, "graj", None)
                        hass.services.call('media_player', 'media_play', {"entity_id": CURR_ENTITIE})
                elif CURR_ENTITIE.startswith('input_boolean.'):
                    curr_state = hass.states.get(CURR_ENTITIE).state
                    if curr_state == 'on':
                        _say_it(hass, "ok, wyłączam", None)
                    if curr_state == 'off':
                        _say_it(hass, "ok, włączam", None)
                    hass.services.call(
                        'input_boolean',
                        'toggle', {
                            "entity_id": CURR_ENTITIE})
                elif CURR_ENTITIE.startswith('switch.'):
                    curr_state = hass.states.get(CURR_ENTITIE).state
                    if curr_state == 'on':
                        _say_it(hass, "ok, wyłączam", None)
                    if curr_state == 'off':
                        _say_it(hass, "ok, włączam", None)
                    if curr_state == 'unavailable':
                        _say_it(hass, "przełącznik jest niedostępny", None)
                    hass.services.call(
                        'switch',
                        'toggle', {
                            "entity_id": CURR_ENTITIE})
                elif CURR_ENTITIE.startswith('light.'):
                    curr_state = hass.states.get(CURR_ENTITIE).state
                    if curr_state == 'on':
                        _say_it(hass, "ok, wyłączam", None)
                    elif curr_state == 'off':
                        _say_it(hass, "ok, włączam", None)
                    elif curr_state == 'unavailable':
                        _say_it(hass, "oświetlnie jest niedostępne", None)
                    hass.services.call(
                        'light',
                        'toggle', {
                            "entity_id": CURR_ENTITIE})
                elif CURR_ENTITIE.startswith('script.'):
                    hass.services.call(
                        'script',
                        CURR_ENTITIE.split('.')[1]
                    )

        else:
            # do some special staff for some entries
            if CURR_ENTITIE == 'sensor.version_info':
                # get the info about upgrade
                state = hass.states.get(CURR_ENTITIE)
                upgrade = state.attributes.get('reinstall_dom_app')
                if upgrade is True:
                    _say_it(
                        hass,
                        "Aktualizuje system do najnowszej wersji. Do usłyszenia.", None)
                    hass.services.call('ais_shell_command', 'execute_upgrade')
                else:
                    _say_it(hass, "Twoja wersja jest aktualna", None)
            else:
                _say_it(hass, "Tej pozycji nie można zmieniać", None)

    if CURR_ENTITIE_ENTERED is True:
        # check if we can change this item
        if can_entity_be_changed(hass, CURR_ENTITIE):
            # these items can be controlled from remote
            # if we are here it means that the enter on the same item was
            # pressed twice, we should do something - to mange the item status
            if CURR_ENTITIE.startswith('input_select.'):
                commit_current_position(hass)
            elif CURR_ENTITIE.startswith('input_number.'):
                commit_current_position(hass)
            elif CURR_ENTITIE.startswith("media_player."):
                # play / pause on selected player
                curr_state = hass.states.get(CURR_ENTITIE).state
                if curr_state == 'playing':
                    if long_press is True:
                        _say_it(hass, "stop", None)
                        hass.services.call('media_player', 'media_stop', {"entity_id": CURR_ENTITIE})
                    else:
                        _say_it(hass, "pauza", None)
                        hass.services.call('media_player', 'media_pause', {"entity_id": CURR_ENTITIE})
                else:
                    _say_it(hass, "graj", None)
                    hass.services.call('media_player', 'media_play', {"entity_id": CURR_ENTITIE})
            elif CURR_ENTITIE.startswith('input_text.'):
                type_to_input_text_from_virtual_keyboard(hass)
        else:
            # eneter on unchanged item
            _say_it(hass, "Tej pozycji nie można zmieniać", None)


def can_entity_be_changed(hass, entity):
    # check if entity can be changed
    if CURR_ENTITIE.startswith((
        "media_player.",
        "input_boolean.",
        "switch.",
        "script.",
        "light.",
        "input_text.",
        "input_select.",
        "input_number."
    )):
        return True
    else:
        return False


def can_entity_be_entered(hass, entity):
    # check if entity can be changed
    if CURR_ENTITIE.startswith((
        "media_player.",
        "input_boolean.",
        "switch.",
        "script.",
        "light."
    )):
        return False
    else:
        return True


def set_on_dpad_down(hass, long_press):
    #
    if CURR_ENTITIE is not None:
        if CURR_ENTITIE.startswith("media_player."):
            # speed up on remote
            state = hass.states.get('input_number.media_player_speed')
            _min = float(state.attributes.get('min'))
            _step = float(state.attributes.get('step'))
            _curr = round(max(float(state.state) - _step, _min), 2)
            _say_it(hass, str(_curr), None)
            _LOGGER.info("speed down the player - info from remote: " + str(_curr))
            hass.services.call(
                'ais_ai_service',
                'publish_command_to_frame', {
                    "key": 'setPlaybackSpeed',
                    "val": _curr
                }
            )
            hass.services.call(
                'input_number',
                'set_value', {
                    "entity_id": "input_number.media_player_speed",
                    "value": _curr})
            return
        elif CURR_ENTITIE.startswith('input_text.') and CURR_ENTITIE_ENTERED:
            set_prev_virtual_keyboard_mode()
            say_curr_virtual_keyboard_mode(hass)


def set_on_dpad_up(hass, long_press):
    #
    if CURR_ENTITIE is not None:
        if CURR_ENTITIE.startswith("media_player."):
            # speed up on remote
            state = hass.states.get('input_number.media_player_speed')
            _max = float(state.attributes.get('max'))
            _step = float(state.attributes.get('step'))
            _curr = round(min(float(state.state) + _step, _max), 2)
            _say_it(hass, str(_curr), None)
            _LOGGER.info("speed up the player - info from remote: " + str(_curr))
            hass.services.call(
                'ais_ai_service',
                'publish_command_to_frame', {
                    "key": 'setPlaybackSpeed',
                    "val": _curr
                }
            )
            hass.services.call(
                'input_number',
                'set_value', {
                    "entity_id": "input_number.media_player_speed",
                    "value": _curr})
            return
        elif CURR_ENTITIE.startswith('input_text.') and CURR_ENTITIE_ENTERED:
            set_next_virtual_keyboard_mode()
            say_curr_virtual_keyboard_mode(hass)


def set_focus_on_prev_entity(hass, long_press):
    # prev on joystick
    if long_press and CURR_ENTITIE is not None:
        if CURR_ENTITIE.startswith("media_player."):
            # seek back on remote
            _LOGGER.info("seek back in the player - info from remote")
            hass.services.call('media_player', 'media_seek', {"entity_id": CURR_ENTITIE, "seek_position": 0})
            return
    # no group is selected go to prev in the groups view menu
    if CURR_GROUP is None:
            set_prev_group_view()
            say_curr_group_view(hass)
            return
    # group is selected
    # check if the entity in the group is selected
    if CURR_ENTITIE is None:
            set_prev_group(hass)
            say_curr_group(hass)
            return
    # entity in the group is selected
    # check if the entity option can be selected
    if can_entity_be_changed(hass, CURR_ENTITIE) and CURR_ENTITIE_ENTERED is True:
        if CURR_ENTITIE.startswith("media_player."):
            hass.services.call('media_player', 'media_previous_track', {"entity_id": CURR_ENTITIE})
        elif CURR_ENTITIE.startswith('input_text.') and CURR_ENTITIE_ENTERED:
            set_prev_virtual_key()
            say_curr_virtual_key(hass)
        else:
            set_prev_position(hass)
    else:
        if CURR_ENTITIE.startswith("media_player.") and CURR_ENTITIE_ENTERED:
            hass.services.call('media_player', 'media_previous_track', {"entity_id": CURR_ENTITIE})
        else:
            # entity not selected or no way to change the entity, go to next one
            set_prev_entity(hass)


def set_focus_on_next_entity(hass, long_press):
    # next on joystick
    if long_press and CURR_ENTITIE is not None:
        if CURR_ENTITIE.startswith("media_player."):
            # seek next on remote
            _LOGGER.info("seek next in the player - info from remote")
            hass.services.call('media_player', 'media_seek', {"entity_id": CURR_ENTITIE, "seek_position": 1})
            return
    # no group is selected go to next in the groups view menu
    if CURR_GROUP is None:
            set_next_group_view()
            say_curr_group_view(hass)
            return
    # group is selected
    # check if the entity in the group is selected
    if CURR_ENTITIE is None:
            set_next_group(hass)
            say_curr_group(hass)
            return
    # entity in the group is selected
    # check if the entity option can be selected
    if can_entity_be_changed(hass, CURR_ENTITIE) and CURR_ENTITIE_ENTERED is True:
        if CURR_ENTITIE.startswith("media_player."):
            hass.services.call('media_player', 'media_next_track', {"entity_id": CURR_ENTITIE})
        elif CURR_ENTITIE.startswith('input_text.') and CURR_ENTITIE_ENTERED:
            set_next_virtual_key()
            say_curr_virtual_key(hass)
        else:
            set_next_position(hass)
    else:
        if CURR_ENTITIE.startswith("media_player.") and CURR_ENTITIE_ENTERED is True:
            hass.services.call('media_player', 'media_next_track', {"entity_id": CURR_ENTITIE})
        else:
            # entity not selected or no way to change the entity, go to next one
            set_next_entity(hass)


def go_up_in_menu(hass):
    # on back on remote
    global CURR_ENTITIE_ENTERED
    # check if the entity in the group is selected
    if CURR_ENTITIE is not None:
        # check if we are browsing files
        if CURR_ENTITIE == "input_select.folder_name":
            # check if we can go up
            state = hass.states.get(CURR_ENTITIE)
            options = state.attributes.get('options')
            if ".." in options:
                _beep_it(hass, 33)
                hass.services.call('ais_drives_service', 'browse_path', {"path": "..", "say": True})
                return
            else:
                # go up in the group menu
                set_curr_group(hass, None)
                say_curr_group(hass)
        elif not CURR_ENTITIE_ENTERED:
            # go up in the group menu
            set_curr_group(hass, None)
            say_curr_group(hass)
        else:
            CURR_ENTITIE_ENTERED = False
            if CURR_ENTITIE.startswith('input_text.'):
                if CURR_VIRTUAL_KEYBOARD_VALUE is None:
                    hass.services.call('input_text', 'set_value', {
                        "entity_id": CURR_ENTITIE, "value": ""})
                else:
                    hass.services.call('input_text', 'set_value', {
                        "entity_id": CURR_ENTITIE, "value": CURR_VIRTUAL_KEYBOARD_VALUE})
            say_curr_entity(hass)
        return
    # no entity is selected, check if the group is selected
    elif CURR_GROUP is not None:
        # go up in the group view menu
        set_curr_group_view()
        say_curr_group_view(hass)
        return
    # can't go up, beep
    _beep_it(hass, 33)


def type_to_input_text(hass, key):
    if CURR_ENTITIE.startswith('input_text.') and CURR_ENTITIE_ENTERED:
        # add the letter to the virtual input
        global CURR_VIRTUAL_KEYBOARD_VALUE
        if CURR_VIRTUAL_KEYBOARD_VALUE is None:
            CURR_VIRTUAL_KEYBOARD_VALUE = chr(key)
        else:
            CURR_VIRTUAL_KEYBOARD_VALUE = CURR_VIRTUAL_KEYBOARD_VALUE + chr(key)

        _say_it(hass, "wpisano: " + chr(key), None)


def type_to_input_text_from_virtual_keyboard(hass):
    # add the letter to the virtual input
    global CURR_VIRTUAL_KEYBOARD_VALUE
    if CURR_VIRTUAL_KEYBOARD_VALUE is None:
        CURR_VIRTUAL_KEYBOARD_VALUE = ""
    if CURR_VIRTUAL_KEY is None:
        if get_curr_virtual_keyboard_mode() == "Usuwanie":
            _say_it(hass, "wybierz tryb usuwania", None)
        else:
            _say_it(hass, "wybierz znak do wpisania", None)
        return

    key = get_curr_virtual_key()
    km = get_curr_virtual_keyboard_mode()
    if km == "Litery":
        key = key.lower()
    if km == "Usuwanie":
        if key == 'ostatni znak':
            text = CURR_VIRTUAL_KEYBOARD_VALUE[:-1]
        elif key == 'ostatni wyraz':
            text = CURR_VIRTUAL_KEYBOARD_VALUE.rsplit(' ', 1)[0]
        else:
            text = ""
    else:
        text = CURR_VIRTUAL_KEYBOARD_VALUE + key

    CURR_VIRTUAL_KEYBOARD_VALUE = text

    text = ""
    if km == "Litery":
        text = "wpisuję literę: " + key.lower()
    elif km == "Wielkie litery":
        text = "wpisuję wielką literę: " + key
    elif km == "Cyfry":
        text = "wpisuję cyfrę: " + key
    elif km == "Znaki specjalne":
        idx = VIRTUAL_KEYBOARD_SYMBOLS.index(key)
        text = "" + VIRTUAL_KEYBOARD_SYMBOLS_NAMES[idx]
        text = "wpisuję znak: " + text
    elif km == "Usuwanie":
        text = "OK, usuwam " + key

    _say_it(hass, text, None)


def go_to_player(hass, say):
    # selecting the player to control via remote
    global CURR_ENTITIE_ENTERED
    if len(GROUP_ENTITIES) == 0:
        get_groups(hass)
    for group in GROUP_ENTITIES:
        if group['entity_id'] == 'group.audio_player':
            set_curr_group(hass, group)
            set_curr_entity(hass, 'media_player.wbudowany_glosnik')
            CURR_ENTITIE_ENTERED = True
            if say:
                _say_it(hass, "Sterowanie odtwarzaczem", None)


def go_home(hass):
    if len(GROUP_ENTITIES) == 0:
        get_groups(hass)
    global CURR_GROUP_VIEW
    CURR_GROUP_VIEW = 'Twój Dom'
    # to reset
    set_curr_group_view()
    say_curr_group_view(hass)


def get_groups(hass):
    global GROUP_ENTITIES
    global ALL_AIS_SENSORS
    entities = hass.states.async_all()
    GROUP_ENTITIES = []

    def add_menu_item(l_entity):
        l_group = {'friendly_name': l_entity.attributes.get('friendly_name'),
                   'order': l_entity.attributes.get('order'),
                   'entity_id': l_entity.entity_id,
                   'entities': l_entity.attributes.get('entity_id'),
                   'context_key_words': l_entity.attributes.get('context_key_words'),
                   'context_answer': l_entity.attributes.get('context_answer'),
                   'context_suffix': l_entity.attributes.get('context_suffix'),
                   'remote_group_view': l_entity.attributes.get('remote_group_view')}
        GROUP_ENTITIES.append(l_group)

    def getKey(item):
        return item['order']

    for entity in entities:
        if entity.entity_id.startswith('group.'):
            remote = entity.attributes.get('remote_group_view')
            if remote is not None:
                if entity.entity_id != 'group.ais_pilot':
                    add_menu_item(entity)
        elif entity.entity_id.startswith('sensor.'):
            # add sensors to the all_ais_sensors group
            device_class = entity.attributes.get('device_class', None)
            if device_class is not None:
                # new_sensor = {
                #     "entity_id": entity.entity_id,
                #     "friendly_name": entity.attributes.get('friendly_name', "Brak nazwy")}
                # ALL_AIS_SENSORS.append(new_sensor)
                # # sort by name
                # entities = []
                # for sorted_sensors in sorted(ALL_AIS_SENSORS, key=lambda i: i['friendly_name']):
                #     entities.append(sorted_sensors['entity_id'])
                # # distinct
                # entities = list(set(entities))
                ALL_AIS_SENSORS.append(entity.entity_id)
                all_unique_sensors = list(set(ALL_AIS_SENSORS))
                all_unique_sensors.sort()
                hass.async_add_job(
                    hass.services.async_call(
                        'group',
                        'set', {
                            "object_id": "all_ais_sensors",
                            "entities": all_unique_sensors
                        }
                    )
                )
                # update sensors on remote
                for group in GROUP_ENTITIES:
                    if group['entity_id'] == 'group.all_ais_sensors':
                        group['entities'] = tuple(all_unique_sensors)

    GROUP_ENTITIES = sorted(GROUP_ENTITIES, key=getKey)


@asyncio.coroutine
async def async_setup(hass, config):
    """Register the process service."""
    warnings.filterwarnings('ignore', module='fuzzywuzzy')
    config = config.get(DOMAIN, {})
    intents = hass.data.get(DOMAIN)
    if intents is None:
        intents = hass.data[DOMAIN] = {}

    for intent_type, utterances in config.get('intents', {}).items():
        conf = intents.get(intent_type)
        if conf is None:
            conf = intents[intent_type] = []
        conf.extend(_create_matcher(utterance) for utterance in utterances)

    @asyncio.coroutine
    async def process(service):
        """Parse text into commands."""
        text = service.data[ATTR_TEXT]
        callback = None
        if 'callback' in service.data:
            callback = service.data['callback']
        await _process(hass, text, callback)

    def process_code(service):
        """Parse remote code into action."""
        text = json.loads(service.data.get(ATTR_TEXT))
        _process_code(hass, text, None)

    def say_it(service):
        """Info to the user."""
        text = service.data[ATTR_TEXT]
        _say_it(hass, text, None)

    def welcome_home(service):
        """Welcome message."""
        text = "Witaj w Domu. Powiedz proszę w czym mogę Ci pomóc?"
        if ais_global.G_OFFLINE_MODE:
            text = "Uwaga, uruchomienie bez dostępu do sieci, część usług może nie działać poprawnie." \
                   "Sprawdź połączenie z Internetem."
        _say_it(hass, text, None)

    @asyncio.coroutine
    def publish_command_to_frame(service):
        key = service.data['key']
        val = service.data['val']
        ip = 'localhost'
        if "ip" in service.data:
            if service.data['ip'] is not None:
                ip = service.data['ip']
        _publish_command_to_frame(hass, key, val, ip)

    def process_command_from_frame(service):
        _process_command_from_frame(hass, service)

    # fix for the problem on box with remote
    def prepare_remote_menu(service):
        get_groups(hass)
        # register context intent
        for menu in GROUP_ENTITIES:
            context_key_words = menu['context_key_words']
            if context_key_words is not None:
                context_key_words = context_key_words.split(',')
                async_register(hass, INTENT_CHANGE_CONTEXT, context_key_words)

    def on_new_iot_device_selection(service):
        iot = service.data['iot'].lower()
        # the name according to the selected model
        if 'dom_' + ais_global.G_MODEL_SONOFF_S20 in iot:
            info = "Inteligentne gniazdo"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_B1 in iot:
            info = "Żarówka"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_TH in iot:
            info = "Przełącznik z czujnikami"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_SLAMPHER in iot:
            info = "Oprawka"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_TOUCH in iot:
            info = "Przełącznik dotykowy"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_POW in iot:
            info = "Przełącznik z pomiarem mocy"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_DUAL in iot:
            info = "Przełącznik podwójny"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_BASIC in iot:
            info = "Przełącznik"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_IFAN in iot:
            info = "Wentylator sufitowy"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_T11 in iot:
            info = "Przełącznik dotykowy pojedynczy"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_T12 in iot:
            info = "Przełącznik dotykowy podwójny"
        elif 'dom_' + ais_global.G_MODEL_SONOFF_T13 in iot:
            info = "Przełącznik dotykowy potrójny"
        else:
            info = "Nowe urządzenie"
        hass.services.call(
            'input_text',
            'set_value', {
                "entity_id": "input_text.ais_iot_device_name",
                "value": info})
        # set the WIFI as an current WIFI (only if empty)
        wifis = hass.states.get('input_select.ais_android_wifi_network')
        if wifis.state == ais_global.G_EMPTY_OPTION and ais_global.GLOBAL_MY_SSID is not None:
            options = wifis.attributes.get('options')
            for o in options:
                if ais_global.GLOBAL_MY_SSID in o:
                    hass.services.call(
                        'input_select',
                        'select_option', {
                            "entity_id": "input_select.ais_android_wifi_network",
                            "option": o})
        # TODO Set the WIFI password if possible

    # register services
    hass.services.async_register(DOMAIN, 'process', process)
    hass.services.async_register(DOMAIN, 'process_code', process_code)
    hass.services.async_register(DOMAIN, 'say_it', say_it)
    hass.services.async_register(DOMAIN, 'welcome_home', welcome_home)
    hass.services.async_register(DOMAIN, 'publish_command_to_frame', publish_command_to_frame)
    hass.services.async_register(DOMAIN, 'process_command_from_frame', process_command_from_frame)
    hass.services.async_register(DOMAIN, 'prepare_remote_menu', prepare_remote_menu)
    hass.services.async_register(DOMAIN, 'on_new_iot_device_selection', on_new_iot_device_selection)

    hass.helpers.intent.async_register(GetTimeIntent())
    hass.helpers.intent.async_register(GetDateIntent())
    hass.helpers.intent.async_register(AisClimateSetTemperature())
    hass.helpers.intent.async_register(AisClimateSetAway())
    hass.helpers.intent.async_register(AisClimateUnSetAway())
    hass.helpers.intent.async_register(AisClimateSetAllOn())
    hass.helpers.intent.async_register(AisClimateSetAllOff())
    hass.helpers.intent.async_register(TurnOnIntent())
    hass.helpers.intent.async_register(TurnOffIntent())
    hass.helpers.intent.async_register(StatusIntent())
    hass.helpers.intent.async_register(PlayRadioIntent())
    hass.helpers.intent.async_register(AisPlayPodcastIntent())
    hass.helpers.intent.async_register(AisPlayYtMusicIntent())
    hass.helpers.intent.async_register(AskQuestionIntent())
    hass.helpers.intent.async_register(ChangeContextIntent())
    hass.helpers.intent.async_register(AisGetWeather())
    hass.helpers.intent.async_register(AisGetWeather48())
    hass.helpers.intent.async_register(AisLampsOn())
    hass.helpers.intent.async_register(AisLampsOff())
    hass.helpers.intent.async_register(AisSwitchesOn())
    hass.helpers.intent.async_register(AisSwitchesOff())
    hass.helpers.intent.async_register(AisOpenCover())
    hass.helpers.intent.async_register(AisCloseCover())
    hass.helpers.intent.async_register(AisStop())
    hass.helpers.intent.async_register(AisPlay())
    hass.helpers.intent.async_register(AisNext())
    hass.helpers.intent.async_register(AisPrev())
    hass.helpers.intent.async_register(AisSceneActive())
    hass.helpers.intent.async_register(AisSayIt())

    async_register(hass, INTENT_GET_WEATHER, [
            '[aktualna] pogoda',
            'pogoda w {location}',
            'pogoda we {location}',
            'jaka jest pogoda',
            'jaka jest pogoda w {location}',
            'jaka jest pogoda we {location}'
    ])
    async_register(hass, INTENT_GET_WEATHER_48, [
            'prognoza pogody',
            'pogoda prognoza',
            'pogoda jutro',
            'pogoda jutro w {location}',
            'pogoda jutro we {location}',
            'jaka będzie pogoda',
            'jaka będzie pogoda w {location}'
    ])
    async_register(hass, INTENT_CLIMATE_SET_TEMPERATURE, [
        "Ustaw temperaturę [ogrzewania] [na] {temp} stopni [w] {item} ",
        "Temperatura ogrzewania {temp} stopni [w] {item}",
        "Temperatura ogrzewania [w] {item} {temp} stopni",
        "Ogrzewanie [w] {item} temperatura {temp} stopni",
        "Ogrzewanie temperatura w {item} {temp} stopni"])
    async_register(hass, INTENT_CLIMATE_SET_AWAY, [
        "Ogrzewanie [na] [w] [tryb] poza domem", "Ogrzewanie włącz [tryb] poza domem"])
    async_register(hass, INTENT_CLIMATE_UNSET_AWAY, [
        "Ogrzewanie [na] [w] [tryb] w domu", "Ogrzewanie wyłącz [tryb] poza domem"])
    async_register(hass, INTENT_CLIMATE_SET_ALL_OFF, ["Wyłącz [całe] ogrzewnie", "Ogrzewanie wyłącz"])
    async_register(hass, INTENT_CLIMATE_SET_ALL_ON, ["Włącz [całe] ogrzewnie", "Ogrzewanie włącz"])
    async_register(hass, INTENT_LAMPS_ON, [
            'włącz światła',
            'zapal światła',
            'włącz wszystkie światła',
            'zapal wszystkie światła'
    ])
    async_register(hass, INTENT_LAMPS_OFF, [
            'zgaś światła',
            'wyłącz światła',
            'wyłącz wszystkie światła',
            'zgaś wszystkie światła'
    ])
    async_register(hass, INTENT_SWITCHES_ON, [
            'włącz przełączniki',
            'włącz wszystkie przełączniki'
    ])
    async_register(hass, INTENT_SWITCHES_OFF, [
            'wyłącz przełączniki',
            'wyłącz wszystkie przełączniki'
    ])
    async_register(hass, INTENT_GET_TIME, [
            'która [jest] [teraz] godzina',
            'którą mamy godzinę',
            'jaki [jest] czas',
            '[jaka] [jest] godzina',
    ])
    async_register(hass, INTENT_GET_DATE, [
            '[jaka] [jest] data',
            'jaki [mamy] [jest] [dzisiaj] dzień',
            'co dzisiaj jest',
            'co [mamy] [jest] dzisiaj',
    ])
    async_register(hass, INTENT_PLAY_RADIO, [
        'Radio {item}',
        'Włącz radio {item}',
        'Graj radio {item}',
        'Graj {item} radio',
        'Przełącz [na] radio [na] {item}',
        'Posłuchał bym radio {item}',
        'Włącz stację radiową {item}',
    ])
    async_register(hass, INTENT_PLAY_PODCAST, [
        'Podcast {item}',
        'Włącz podcast {item}',
        'Graj podcast {item}',
        'Graj {item} podcast',
        'Przełącz [na] podcast {item}',
        'Posłuchał bym podcast {item}',
    ])
    async_register(hass, INTENT_PLAY_YT_MUSIC, [
        'Muzyka {item}',
        'Włącz muzykę {item}',
        'Graj muzykę {item}',
        'Graj {item} muzykę',
        'Przełącz [na] muzykę [na] {item}',
        'Posłuchał bym muzykę {item}',
        'Włącz [z] [na] YouTube {item}',
        'YouTube {item}'
    ])
    async_register(hass, INTENT_TURN_ON,
                   ['Włącz {item}', 'Zapal światło w {item}'])
    async_register(hass, INTENT_TURN_OFF,
                   ['Wyłącz {item}', 'Zgaś Światło w {item}'])
    async_register(hass, INTENT_STATUS, [
        'Jaka jest {item}', 'Jaki jest {item}',
        'Jak jest {item}', 'Jakie jest {item}',
        '[jaki] [ma] status {item}'])
    async_register(hass, INTENT_ASK_QUESTION, [
        'Co to jest {item}', 'Kto to jest {item}',
        'Znajdź informację o {item}', 'Znajdź informacje o {item}',
        'Wyszukaj informację o {item}', 'Wyszukaj informacje o {item}',
        'Wyszukaj {item}', 'Kim jest {item}', 'Informacje o {item}',
        'Czym jest {item}', 'Opowiedz mi o {intem}',
        'Informację na temat {item}', 'Co wiesz o {item}',
        'Co wiesz na temat {item}', 'Opowiedz o {item}',
        'Kim są {item}', 'Kto to {item}'])
    async_register(hass, INTENT_OPEN_COVER, ['Otwórz {item}', 'Odsłoń {item}'])
    async_register(hass, INTENT_CLOSE_COVER, ['Zamknij {item}', 'Zasłoń {item}'])
    async_register(hass, INTENT_STOP, ['Stop', 'Zatrzymaj', 'Koniec', 'Pauza', 'Zaniechaj', 'Stój'])
    async_register(hass, INTENT_PLAY, ['Start', 'Graj', 'Odtwarzaj'])
    async_register(hass, INTENT_SCENE, ['Scena {item}', 'Aktywuj [scenę] {item}'])
    async_register(hass, INTENT_NEXT, ['[włącz] następny', '[włącz] kolejny', '[graj] następny', '[graj] kolejny'])
    async_register(hass, INTENT_PREV, ['[włącz] poprzedni', '[włącz] wcześniejszy', '[graj] poprzedni',
                                       '[graj] wcześniejszy'])
    async_register(hass, INTENT_SAY_IT, ['Powiedz', 'Mów', 'Powiedz {item}', 'Mów {item}', 'Echo {item}'])

    return True


def _publish_command_to_frame(hass, key, val, ip):
    # sent the command to the android frame via http
    url = G_HTTP_REST_SERVICE_BASE_URL.format(ip)

    if key == "WifiConnectToSid":
        # enable the wifi info
        hass.async_run_job(
            hass.services.async_call(
                'input_boolean',
                'input_boolean',
                'turn_on', {"entity_id": "input_boolean.ais_android_wifi_changes_notify"})
        )
        ssid = val.split(';')[0]
        if ssid is None or ssid == "-" or ssid == "":
            _say_it(hass, "Wybierz sieć WiFi z listy", None)
            return


        # TODO get password from file
        password = hass.states.get('input_text.ais_android_wifi_password').state
        if len(password.strip()) == 0:
            _say_it(hass, "ok, przełączam na sieć: " + ssid, None)
        else:
            _say_it(hass, "ok, łączę z siecią: " + ssid, None)

        wifi_type = val.split(';')[-3]
        bssid = val.split(';')[-1].replace("MAC:", "").strip()
        requests.post(
            url + '/command',
            json={key: ssid, "ip": ip, "WifiNetworkPass": password, "WifiNetworkType": wifi_type, "bssid": bssid},
            timeout=2)

    elif key == "WifiConnectTheDevice":
        iot = val.split(';')[0]
        if iot == ais_global.G_EMPTY_OPTION:
            _say_it(hass, "wybierz wifi do której mam dołączyć urządzenie", None)
            return
        # check if wifi is selected
        ssid = hass.states.get('input_select.ais_android_wifi_network').state.split(';')[0]
        if ssid == ais_global.G_EMPTY_OPTION:
            _say_it(hass, "wybierz wifi do której mam dołączyć urządzenie", None)
            return

        # take bssid
        bssid = val.split(';')[-1].replace("MAC:", "").strip()

        # check the frequency
        wifi_frequency_mhz = val.split(';')[-2]
        if not wifi_frequency_mhz.startswith("2.4"):
            _say_it(hass, "Urządzenia mogą pracować tylko w sieci 2.4 GHz, wybierz inną sieć.", None)

        # check if name is selected, if not then add the device name
        name = hass.states.get('input_text.ais_iot_device_name').state
        # friendly name (32 chars max)
        if name == "":
            name = iot
        if len(name) > 32:
            _say_it(hass, "nazwa urządzenie może mieć maksymalnie 32 znaki", None)
            return
        _say_it(hass, "dodajemy: " + name, None)
        password = hass.states.get('input_text.ais_iot_device_wifi_password').state

        requests.post(
            url + '/command',
            json={key: iot, "ip": ip, "WifiNetworkPass": password, "WifiNetworkSsid": ssid,
                  "IotName": name, "bsssid": bssid},
            timeout=2)
    else:
        try:
            requests.post(
                url + '/command',
                json={key: val, "ip": ip},
                timeout=2)
        except Exception as e:
            _LOGGER.info(
                "_publish_command_to_frame requests.post problem key: "
                + str(key) + " val: " + str(val) + " ip " + str(ip) + str(e))


def _wifi_rssi_to_info(rssi):
    info = "moc nieznana"
    if rssi > -31:
        return "moc doskonała (" + str(rssi) + ")"
    if rssi > -68:
        return "moc bardzo dobra (" + str(rssi) + ")"
    if rssi > -71:
        return "moc dobra (" + str(rssi) + ")"
    if rssi > -81:
        return "moc słaba (" + str(rssi) + ")"
    if rssi > -91:
        return "moc bardzo słaba (" + str(rssi) + ")"
    return info


def _wifi_frequency_info(mhz):
    if str(mhz).startswith("2"):
        return "2.4 GHz"
    elif str(mhz).startswith("5"):
        return "5 GHz"
    return str(mhz)


def _publish_wifi_status(hass, service):
    wifis = json.loads(service.data["payload"])
    wifis_names = [ais_global.G_EMPTY_OPTION]
    for item in wifis["ScanResult"]:
        if len(item["ssid"]) > 0:
            wifis_names.append(
                item["ssid"] + "; " +
                _wifi_rssi_to_info(item["rssi"]) +
                "; " + item["capabilities"] + "; " + _wifi_frequency_info(item["frequency_mhz"]) +
                "; MAC: " + item["bssid"])
    hass.async_run_job(
        hass.services.call(
            'input_select',
            'set_options', {
                "entity_id": "input_select.ais_android_wifi_network",
                "options": wifis_names})
    )
    return len(wifis_names)-1


def _process_command_from_frame(hass, service):
    # process from frame
    _LOGGER.info('_process_command_from_frame: ' + str(service.data))
    if "web_hook_json" in service.data:
        import ast
        service.data = ast.literal_eval(service.data["web_hook_json"])
    callback = None
    if "callback" in service.data:
        callback = service.data["callback"]
    if service.data["topic"] == 'ais/speech_command':
        hass.async_run_job(
            _process(hass, service.data["payload"], callback)
        )
        return
    elif service.data["topic"] == 'ais/key_command':
        _process_code(hass, json.loads(service.data["payload"]), callback)
        return
    elif service.data["topic"] == 'ais/speech_text':
        _say_it(hass, service.data["payload"], callback)
        return
    elif service.data["topic"] == 'ais/speech_status':
        # TODO pause/resume, service.data["payload"] can be: START -> DONE/ERROR
        _LOGGER.info('speech_status: ' + str(service.data["payload"]))
        return
    elif service.data["topic"] == 'ais/add_bookmark':
        _LOGGER.info('add_bookmark: ' + str(service.data["payload"]))
        try:
            bookmark = json.loads(service.data["payload"])
            if bookmark["media_source"] != ais_global.G_AN_LOCAL:
                return
            hass.async_run_job(
                hass.services.call('ais_bookmarks',
                                   'add_bookmark',
                                   {"attr":
                                    {"media_title": bookmark["media_title"],
                                     "source": bookmark["media_source"],
                                     "media_position": bookmark["media_position"],
                                     "media_content_id": bookmark["media_content_id"],
                                     "media_stream_image": bookmark["media_stream_image"]}})
            )
        except Exception as e:
            _LOGGER.info("problem to add_bookmark: " + str(e))
        return
    elif service.data["topic"] == 'ais/player_speed':
        speed = json.loads(service.data["payload"])
        # _say_it(hass, "prędkość odtwarzania: " + str(speed["currentSpeed"]), callback)
        # hass.services.call(
        #     'input_number',
        #     'set_value', {
        #         "entity_id": "input_number.media_player_speed",
        #         "value": round(speed["currentSpeed"], 2)})
        return
    elif service.data["topic"] == 'ais/wifi_scan_info':
        len_wifis = _publish_wifi_status(hass, service)
        info = "Mamy dostępne " + str(len_wifis) + " wifi."
        _say_it(hass, info, callback)
        return
    elif service.data["topic"] == 'ais/iot_scan_info':
        iot = json.loads(service.data["payload"])
        iot_names = [ais_global.G_EMPTY_OPTION]
        for item in iot["ScanResult"]:
            if len(item["ssid"]) > 0:
                iot_names.append(
                    item["ssid"] + "; " +
                    _wifi_rssi_to_info(item["rssi"]) +
                    "; " + item["capabilities"])
        hass.async_run_job(
            hass.services.call(
                'input_select',
                'set_options', {
                    "entity_id": "input_select.ais_iot_devices_in_network",
                    "options": iot_names})
        )
        if len(iot_names) == 1:
            info = "Nie znaleziono żadnego nowego urządzenia"
        elif len(iot_names) == 2:
            if item["model"] == ais_global.G_MODEL_SONOFF_S20:
                info = "Znaleziono nowe inteligentne gniazdo"
            elif item["model"] == ais_global.G_MODEL_SONOFF_SLAMPHER:
                info = "Znaleziono nową oprawkę"
            elif item["model"] == ais_global.G_MODEL_SONOFF_TOUCH:
                info = "Znaleziono nowy przełącznik dotykowy"
            elif item["model"] == ais_global.G_MODEL_SONOFF_TH:
                info = "Znaleziono nowy przełącznik z czujnikami"
            elif item["model"] == ais_global.G_MODEL_SONOFF_B1:
                info = "Znaleziono nową żarówkę"
            elif item["model"] == ais_global.G_MODEL_SONOFF_POW:
                info = "Znaleziono nowy przełącznik z pomiarem mocy"
            elif item["model"] == ais_global.G_MODEL_SONOFF_DUAL:
                info = "Znaleziono nowy podwójny przełącznik"
            elif item["model"] == ais_global.G_MODEL_SONOFF_BASIC:
                info = "Znaleziono nowy przełącznik"
            elif item["model"] == ais_global.G_MODEL_SONOFF_IFAN:
                info = "Znaleziono nowy wentylator sufitowy"
            elif item["model"] == ais_global.G_MODEL_SONOFF_T11:
                info = "Znaleziono nowy przełącznik dotykowy pojedynczy"
            elif item["model"] == ais_global.G_MODEL_SONOFF_T12:
                info = "Znaleziono nowy przełącznik dotykowy podwójny"
            elif item["model"] == ais_global.G_MODEL_SONOFF_T13:
                info = "Znaleziono nowy przełącznik dotykowy potrójny"
            else:
                info = "Znaleziono nowe urządzenie, to urządzenie może"
                info += " nie być w pełni wspierane przez system 'Asystent domowy'"
        else:
            info = "Znaleziono " + str(len(iot_names)-1) + " nowe urządzenia."
        _say_it(hass, info, callback)
        return
    elif service.data["topic"] == 'ais/wifi_status_info':
        _publish_wifi_status(hass, service)
        return
    elif service.data["topic"] == 'ais/wifi_connection_info':
        # current connection info
        cci = json.loads(service.data["payload"])
        info = "Połączenie Wifi: "
        if "ssid" in cci:
            ais_global.set_my_ssid(cci["ssid"])
            if cci["ssid"] == "<unknown ssid>":
                info += "brak połączenia"
            else:
                info += cci["ssid"]
                if "link_speed_mbps" in cci:
                    info += "; prędkość: " + str(cci["link_speed_mbps"]) + " megabitów na sekundę"
                if "rssi" in cci:
                    info += "; " + _wifi_rssi_to_info(cci["rssi"])
                if "frequency_mhz" in cci:
                    info += "; " + _wifi_frequency_info(cci["frequency_mhz"])

        hass.states.async_set(
            'sensor.ais_android_wifi_current_network_info',
            info, {'custom_ui_state_card': "state-card-text"}
            )
        return
    elif service.data["topic"] == 'ais/wifi_state_change_info':
        # current connection info
        cci = json.loads(service.data["payload"])
        ais_global.set_my_ssid(cci["ssid"])
        info = "Wifi: "
        if "ssid" in cci:
            if 'dom_' + ais_global.G_MODEL_SONOFF_S20 in cci["ssid"].lower():
                info += "gniazdo "
            elif 'dom_' + ais_global.G_MODEL_SONOFF_B1 in cci["ssid"].lower():
                info += "żarówka "
            elif 'dom_' + ais_global.G_MODEL_SONOFF_TH in cci["ssid"].lower():
                info += "przełącznik z czujnikami "
            elif 'dom_' + ais_global.G_MODEL_SONOFF_SLAMPHER in cci["ssid"].lower():
                info += "oprawka "
            elif 'dom_' + ais_global.G_MODEL_SONOFF_TOUCH in cci["ssid"].lower():
                info += "przełącznik dotykowy "
            elif 'dom_' + ais_global.G_MODEL_SONOFF_POW in cci["ssid"].lower():
                info += "przełącznik z pomiarem mocy"
            elif 'dom_' + ais_global.G_MODEL_SONOFF_DUAL in cci["ssid"].lower():
                info += "podwójny przełącznik"
            elif 'dom_' + ais_global.G_MODEL_SONOFF_BASIC in cci["ssid"].lower():
                info += "przełącznik"
            elif 'dom_' + ais_global.G_MODEL_SONOFF_IFAN in cci["ssid"].lower():
                info += "wentylator sufitowy"
            elif 'dom_' + ais_global.G_MODEL_SONOFF_T11 in cci["ssid"].lower():
                info = "przełącznik dotykowy pojedynczy"
            elif 'dom_' + ais_global.G_MODEL_SONOFF_T12 in cci["ssid"].lower():
                info = "przełącznik dotykowy podwójny"
            elif 'dom_' + ais_global.G_MODEL_SONOFF_T13 in cci["ssid"].lower():
                info = "przełącznik dotykowy potrójny"
            else:
                info += cci["ssid"] + " "
        if "state" in cci:
            info += cci["state"]
        # check if we are now online
        if ais_global.GLOBAL_MY_IP == "127.0.0.1":
            ais_global.set_global_my_ip()
            if ais_global.GLOBAL_MY_IP != "127.0.0.1":
                pass
                # if yes then try to reload the cloud and other components
                # TODO reload invalid components
                # hass.async_run_job(async_load_platform(hass, 'sun', 'sun', {}, {}))
                # hass.async_run_job(async_load_platform(hass, 'ais_cloud', 'ais_cloud', {}, {}))
                # hass.async_run_job(async_load_platform(hass, 'ais_yt_service', 'ais_yt_service', {}, {}))
                # hass.async_run_job(async_load_platform(hass, 'ais_knowledge_service', 'ais_knowledge_service'...
        state = hass.states.get('input_boolean.ais_android_wifi_changes_notify').state
        if state == 'on':
            _say_it(hass, info, callback)
        return
    elif service.data["topic"] == 'ais/go_to_player':
        go_to_player(hass, False)
    else:
        # TODO process this without mqtt
        # player_status and speech_status
        mqtt.async_publish(
            hass, service.data["topic"], service.data["payload"], 2)
        # TODO
    return


def _post_message(message, hosts):
    """Post the message to TTS service."""
    message = message.replace("°C", "stopni Celsjusza")
    j_data = {
            "text": message,
            "pitch": ais_global.GLOBAL_TTS_PITCH,
            "rate": ais_global.GLOBAL_TTS_RATE,
            "voice": ais_global.GLOBAL_TTS_VOICE
            }
    for h in hosts:
        _LOGGER.info("sending text_to_speech to: " + str(h))
        try:
            requests.post(
                G_HTTP_REST_SERVICE_BASE_URL.format(h) + '/text_to_speech',
                json=j_data,
                timeout=1
            )
        except Exception as e:
            _LOGGER.info("problem to send the text to speech via http: " + str(e))


def _beep_it(hass, tone):
    """Post the beep to Android frame."""
    # https://android.googlesource.com/platform/frameworks/base/+/b267554/media/java/android/media/ToneGenerator.java
    hass.services.call(
        'ais_ai_service',
        'publish_command_to_frame', {
            "key": 'tone',
            "val": tone
        }
    )


def _say_it(hass, message, caller_ip=None):
    # sent the tts message to the panel via http api
    global GLOBAL_TTS_TEXT
    l_hosts = ['localhost']

    # check if we should inform other speaker
    player_name = hass.states.get('input_select.tts_player').state
    device_ip = None
    if player_name is not None:
        device = ais_cloud.get_player_data(player_name)
        if device is not None:
            device_ip = device["device_ip"]
            if device_ip not in ['localhost', '127.0.0.1']:
                l_hosts.append(device_ip)

    # check if we should inform back the caller speaker
    # the local caller has ip like 192.168.1.45
    # internal_ip = hass.states.get('sensor.internal_ip_address').state
    if ais_global.GLOBAL_MY_IP is None:
        ais_global.set_global_my_ip()
    if caller_ip is not None:
        if caller_ip not in ['localhost', '127.0.0.1', device_ip, ais_global.GLOBAL_MY_IP]:
            l_hosts.append(caller_ip)

    _post_message(message, l_hosts)

    if len(message) > 199:
        GLOBAL_TTS_TEXT = message[0: 199] + '...'
    else:
        GLOBAL_TTS_TEXT = message
    hass.states.async_set(
        'sensor.ais_knowledge_answer', 'ok', {
            'custom_ui_state_card': 'state-card-text',
            'text': GLOBAL_TTS_TEXT
            })


def _create_matcher(utterance):
    """Create a regex that matches the utterance."""
    # Split utterance into parts that are type: NORMAL, GROUP or OPTIONAL
    # Pattern matches (GROUP|OPTIONAL): Change light to [the color] {item}
    parts = re.split(r'({\w+}|\[[\w\s]+\] *)', utterance)
    # Pattern to extract name from GROUP part. Matches {item}
    group_matcher = re.compile(r'{(\w+)}')
    # Pattern to extract text from OPTIONAL part. Matches [the color]
    optional_matcher = re.compile(r'\[([\w ]+)\] *')

    pattern = ['^']
    for part in parts:
        group_match = group_matcher.match(part)
        optional_match = optional_matcher.match(part)

        # Normal part
        if group_match is None and optional_match is None:
            pattern.append(part)
            continue

        # Group part
        if group_match is not None:
            pattern.append(
                r'(?P<{}>[\w ]+?)\s*'.format(group_match.groups()[0]))

        # Optional part
        elif optional_match is not None:
            pattern.append(r'(?:{} *)?'.format(optional_match.groups()[0]))

    pattern.append('$')
    return re.compile(''.join(pattern), re.I)


def _process_code(hass, data, callback):
    """Process a code from remote."""
    global CURR_BUTTON_CODE
    global CURR_BUTTON_LONG_PRESS
    global CURR_ENTITIE_ENTERED
    if 'Action' not in data or 'KeyCode' not in data:
        return
    # check if we have callback
    if callback is not None:
        # TODO the answer should go to the correct clien
        pass

    action = data["Action"]
    code = data["KeyCode"]

    # fix - when the mouse mode on remote is on, the remote is sending only the code 23 (OK) as key down (action 0)
    # to handle this we are ignoring the key up (action 1), and key down (action 0) is changing to key up (action 1)
    if code == 23:
        if action == 1:
            return
        else:
            action = 1

    # ACTION_DOWN = 0; ACTION_UP = 1;
    if action == 0:
        CURR_BUTTON_LONG_PRESS = False
        if "RepeatCount" in data:
            if data["RepeatCount"] > 0:
                CURR_BUTTON_LONG_PRESS = True
        if CURR_BUTTON_LONG_PRESS is False:
            return

    elif action == 2:
        # ACTION_MULTIPLE = 2;
        # check if long press will be send to us
        _LOGGER.info("long press on " + str(data))
        return

    elif action == 1:
        # ACTION_UP
        # to prevent up action after long press
        if CURR_BUTTON_LONG_PRESS is True:
            CURR_BUTTON_LONG_PRESS = False
            return

    _LOGGER.info("KeyCode: -> " + str(code))
    # set the code in global variable
    CURR_BUTTON_CODE = code
    # show the code in web app
    hass.states.set('binary_sensor.ais_remote_button', code)
    # decode Key Events
    # codes according to android.view.KeyEvent
    if code == 93:
        # PG- -> KEYCODE_PAGE_DOWN
        set_bookmarks_curr_group(hass)
        set_curr_entity(hass, 'input_select.ais_bookmark_last_played')
        CURR_ENTITIE_ENTERED = True
        say_curr_entity(hass)
    elif code == 92:
        # PG+ -> KEYCODE_PAGE_UP
        set_bookmarks_curr_group(hass)
        CURR_ENTITIE_ENTERED = True
        set_curr_entity(hass, 'input_select.ais_bookmark_favorites')

        say_curr_entity(hass)
    elif code == 4:
        # Back arrow, go up in menu/groups -> KEYCODE_BACK
        # or go up in local folder structure
        go_up_in_menu(hass)
    elif code == 82:
        # Menu -> KEYCODE_MENU
        set_next_group_view()
        say_curr_group_view(hass)
    elif code == 164:
        # Mute -> KEYCODE_VOLUME_MUTE
        pass
    elif code == 71:
        # MIC DOWN -> KEYCODE_LEFT_BRACKET
        pass
    elif code == 72:
        # MIC UP -> KEYCODE_RIGHT_BRACKET
        pass
    elif code == 19:
        # Dpad up -> KEYCODE_DPAD_UP
        set_on_dpad_up(hass, CURR_BUTTON_LONG_PRESS)
        pass
    elif code == 20:
        # Dpad down -> KEYCODE_DPAD_DOWN
        set_on_dpad_down(hass, CURR_BUTTON_LONG_PRESS)
        pass
    elif code == 21:
        # Dpad left -> KEYCODE_DPAD_LEFT
        set_focus_on_prev_entity(hass, CURR_BUTTON_LONG_PRESS)
    elif code == 22:
        # Dpad right -> KEYCODE_DPAD_RIGHT
        set_focus_on_next_entity(hass, CURR_BUTTON_LONG_PRESS)
    elif code == 23:
        # Dpad center -> KEYCODE_DPAD_CENTER
        select_entity(hass, CURR_BUTTON_LONG_PRESS)
    elif code == 25:
        # Volume down -> KEYCODE_VOLUME_DOWN
        pass
    elif code == 24:
        # Volume up -> KEYCODE_VOLUME_UP
        pass
    elif code == 190:
        # go home -> KEYCODE_HOME
        if CURR_BUTTON_LONG_PRESS:
            go_to_player(hass, True)
        else:
            go_home(hass)
    # other code on text field
    else:
        type_to_input_text(hass, code)


@asyncio.coroutine
def _process(hass, text, callback):
    """Process a line of text."""
    _LOGGER.info('Process text: ' + text)
    # clear text
    text = text.replace("&", " and ").lower()
    global CURR_BUTTON_CODE
    s = False
    m = None
    m_org = None
    found_intent = None
    # first check the conversation intents
    conv_intents = hass.data.get('conversation', {})
    for intent_type, matchers in conv_intents.items():
        for matcher in matchers:
            match = matcher.match(text)
            if not match:
                continue
            response = yield from hass.helpers.intent.async_handle(
                'conversation', intent_type,
                {key: {'value': value} for key, value
                 in match.groupdict().items()}, text)
            return response

    # check the AIS dom intents
    intents = hass.data.get(DOMAIN, {})
    try:
        for intent_type, matchers in intents.items():
            if found_intent is not None:
                break
            for matcher in matchers:
                match = matcher.match(text)
                if match:
                    # we have a match
                    found_intent = intent_type
                    m, s = yield from hass.helpers.intent.async_handle(
                        DOMAIN, intent_type,
                        {key: {'value': value} for key, value
                         in match.groupdict().items()}, text)
                    break
        # the item was match as INTENT_TURN_ON but we don't have such device - maybe it is radio or podcast???
        if s is False and found_intent == INTENT_TURN_ON:
            m_org = m
            m, s = yield from hass.helpers.intent.async_handle(
                DOMAIN, INTENT_PLAY_RADIO,
                {key: {'value': value} for key, value
                 in match.groupdict().items()}, text.replace("włącz", "włącz radio"))
            if s is False:
                m, s = yield from hass.helpers.intent.async_handle(
                    DOMAIN, INTENT_PLAY_PODCAST,
                    {key: {'value': value} for key, value
                     in match.groupdict().items()}, text.replace("włącz", "włącz podcast"))
            if s is False:
                m = m_org
        # the item was match as INTENT_TURN_ON but we don't have such device - maybe it is climate???
        if s is False and found_intent == INTENT_TURN_ON:
            m_org = m
            m, s = yield from hass.helpers.intent.async_handle(
                DOMAIN, INTENT_CLIMATE_SET_ALL_ON,
                {key: {'value': value} for key, value
                    in match.groupdict().items()}, text)
            if s is False:
                m = m_org
        # the item was match as INTENT_TURN_OFF but we don't have such device - maybe it is climate???
        if s is False and found_intent == INTENT_TURN_OFF:
            m_org = m
            m, s = yield from hass.helpers.intent.async_handle(
                DOMAIN, INTENT_CLIMATE_SET_ALL_OFF,
                {key: {'value': value} for key, value
                    in match.groupdict().items()}, text)
            if s is False:
                m = m_org
        # the was no match - try again but with current context
        if found_intent is None:
            suffix = GROUP_ENTITIES[get_curr_group_idx()]['context_suffix']
            if suffix is not None:
                for intent_type, matchers in intents.items():
                    if found_intent is not None:
                        break
                    for matcher in matchers:
                        match = matcher.match(suffix + " " + text)
                        if match:
                            # we have a match
                            found_intent = intent_type
                            m, s = yield from hass.helpers.intent.async_handle(
                                DOMAIN, intent_type,
                                {key: {'value': value} for key, value
                                 in match.groupdict().items()},
                                suffix + " " + text)
                            # reset the curr button code
                            # TODO the mic should send a button code too
                            # in this case we will know if the call source
                            CURR_BUTTON_CODE = 0
                            break

        # the was no match - try again but with player context
        # we should get media player source first
        if found_intent is None:
            if CURR_ENTITIE == 'media_player.wbudowany_glosnik' and CURR_ENTITIE_ENTERED:
                state = hass.states.get(CURR_ENTITIE)
                if "source" in state.attributes:
                    suffix = ""
                    source = state.attributes.get('source')
                    if source == ais_global.G_AN_MUSIC:
                        suffix = 'youtube'
                    elif source == ais_global.G_AN_RADIO:
                        suffix = 'radio'
                    elif source == ais_global.G_AN_PODCAST:
                        suffix = 'podcast'
                    if suffix != "":
                        for intent_type, matchers in intents.items():
                            if found_intent is not None:
                                break
                            for matcher in matchers:
                                match = matcher.match(suffix + " " + text)
                                if match:
                                    # we have a match
                                    found_intent = intent_type
                                    m, s = yield from hass.helpers.intent.async_handle(
                                        DOMAIN, intent_type,
                                        {key: {'value': value} for key, value
                                         in match.groupdict().items()},
                                        suffix + " " + text)
                                    # reset the curr button code
                                    CURR_BUTTON_CODE = 0
                                    break
        if s is False or found_intent is None:
            # no success - try to ask the cloud
            if m is None:
                # no message / no match
                m = 'Nie rozumiem ' + text
            # asking without the suffix
            ws_resp = aisCloudWS.ask(text, m)
            m = ws_resp.text

    except Exception as e:
        _LOGGER.warning('_process: ' + str(e))
        m = "Przepraszam, ale mam problem ze zrozumieniem: " + text
    # return response to the ais dom
    if m != 'DO_NOT_SAY':
        _say_it(hass, m, callback)
    # return response to the hass conversation
    intent_resp = intent.IntentResponse()
    intent_resp.async_set_speech(m)
    intent_resp.hass = hass
    return intent_resp


@core.callback
def _match_entity(hass, name):
    """Match a name to an entity."""
    from fuzzywuzzy import process as fuzzyExtract
    entities = {state.entity_id: state.name for state
                in hass.states.async_all()}
    try:
        entity_id = fuzzyExtract.extractOne(
            name, entities, score_cutoff=86)[2]
    except Exception as e:
        entity_id = None

    if entity_id is not None:
        return hass.states.get(entity_id)
    else:
        return None


class TurnOnIntent(intent.IntentHandler):
    """Handle turning item on intents."""

    intent_type = INTENT_TURN_ON
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle turn on intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        name = slots['item']['value']
        entity = _match_entity(hass, name)
        success = False

        if not entity:
            message = 'Nie znajduję urządzenia do włączenia, o nazwie: ' + name
        else:
            # check if we can turn_on on this device
            if isSwitch(entity.entity_id):
                assumed_state = entity.attributes.get(ATTR_ASSUMED_STATE, False)
                if assumed_state is False:
                    if entity.state == 'on':
                        # check if the device is already on
                        message = 'Urządzenie ' + name + ' jest już włączone'
                    elif entity.state == 'unavailable':
                        message = 'Urządzenie ' + name + ' jest niedostępne'
                    else:
                        assumed_state = True
                if assumed_state:
                    yield from hass.services.async_call(
                        core.DOMAIN, SERVICE_TURN_ON, {
                            ATTR_ENTITY_ID: entity.entity_id,
                        }, blocking=True)
                    message = 'OK, włączono {}'.format(entity.name)
                success = True
            else:
                message = 'Urządzenia ' + name + ' nie można włączyć'
        return message, success


class TurnOffIntent(intent.IntentHandler):
    """Handle turning item off intents."""

    intent_type = INTENT_TURN_OFF
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle turn off intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        name = slots['item']['value']
        entity = _match_entity(hass, name)
        success = False
        if not entity:
            msg = 'Nie znajduję urządzenia do wyłączenia, o nazwie: ' + name
        else:
            # check if we can turn_off on this device
            if isSwitch(entity.entity_id):
                assumed_state = entity.attributes.get(ATTR_ASSUMED_STATE, False)
                if assumed_state is False:
                    # check if the device is already off
                    if entity.state == 'off':
                        msg = 'Urządzenie {} jest już wyłączone'.format(
                            entity.name)
                    elif entity.state == 'unavailable':
                        msg = 'Urządzenie {}} jest niedostępne'.format(
                            entity.name)
                    else:
                        assumed_state = True
                if assumed_state:
                    yield from hass.services.async_call(
                        core.DOMAIN, SERVICE_TURN_OFF, {
                            ATTR_ENTITY_ID: entity.entity_id,
                        }, blocking=True)
                    msg = 'OK, wyłączono {}'.format(entity.name)
                    success = True
            else:
                msg = 'Urządzenia ' + name + ' nie można wyłączyć'
        return msg, success


class StatusIntent(intent.IntentHandler):
    """Handle status item on intents."""

    intent_type = INTENT_STATUS
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle status intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        name = slots['item']['value']
        entity = _match_entity(hass, name)
        success = False

        if not entity:
            message = 'Nie znajduję informacji o: ' + name
            success = False
        else:
            unit = entity.attributes.get(ATTR_UNIT_OF_MEASUREMENT)
            state = entity.state
            state = translate_state(state)
            if unit is None:
                value = state
            else:
                value = "{} {}".format(state, unit)
            message = format(entity.name) + ': ' + value
            success = True
        return message, success


class PlayRadioIntent(intent.IntentHandler):
    """Handle PlayRadio intents."""

    intent_type = INTENT_PLAY_RADIO
    slot_schema = {
        'item': cv.string
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        item = slots['item']['value']
        station = item
        success = False
        if not station:
            message = 'Nie wiem jaką stację chcesz włączyć.'
        else:
            ws_resp = aisCloudWS.audio(
                station, ais_global.G_AN_RADIO, intent_obj.text_input)
            json_ws_resp = ws_resp.json()
            json_ws_resp["audio_type"] = ais_global.G_AN_RADIO
            name = json_ws_resp['name']
            if len(name.replace(" ", "")) == 0:
                message = "Niestety nie znajduję radia " + station
            else:
                yield from hass.services.async_call(
                     'ais_cloud', 'play_audio',
                     json_ws_resp, blocking=False)
                message = "OK, gramy radio " + name
                success = True
        return message, success


class AisPlayPodcastIntent(intent.IntentHandler):
    """Handle Podcast intents."""

    intent_type = INTENT_PLAY_PODCAST
    slot_schema = {
        'item': cv.string
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        item = slots['item']['value']
        success = False

        if not item:
            message = 'Nie wiem jaką audycję chcesz posłuchać.'
        else:
            ws_resp = aisCloudWS.audio(
                item, ais_global.G_AN_PODCAST, intent_obj.text_input)
            json_ws_resp = ws_resp.json()
            json_ws_resp["audio_type"] = ais_global.G_AN_PODCAST
            name = json_ws_resp['name']
            if len(name.replace(" ", "")) == 0:
                message = "Niestety nie znajduję podcasta " + item
            else:
                yield from hass.services.async_call(
                     'ais_cloud', 'play_audio',
                     json_ws_resp, blocking=False)
                message = "OK, pobieram odcinki audycji " + item
                success = True
        return message, success


class AisPlayYtMusicIntent(intent.IntentHandler):
    """Handle Music intents."""

    intent_type = INTENT_PLAY_YT_MUSIC
    slot_schema = {
        'item': cv.string
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        item = slots['item']['value']
        success = False

        if not item:
            message = 'Nie wiem jaką muzykę mam szukać '
        else:
            success = True
            yt_query = {}
            yt_query["audio_type"] = ais_global.G_AN_MUSIC
            yt_query["text"] = item
            yield from hass.services.async_call(
                 'ais_cloud', 'play_audio',
                 yt_query, blocking=False)
            message = "OK, szukam na YouTube " + item
            success = True
        return message, success


class AskQuestionIntent(intent.IntentHandler):
    """Handle AskQuestion intents."""

    intent_type = INTENT_ASK_QUESTION
    slot_schema = {
        'item': cv.string
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        item = slots['item']['value']
        question = item
        if not question:
            message = 'Nie wiem o co zapytać, ' + question
            return message, False
        else:
            yield from hass.services.async_call(
                 'ais_knowledge_service',
                 'ask', {"text": question, "say_it": True}, blocking=True)
        return 'DO_NOT_SAY', True


class ChangeContextIntent(intent.IntentHandler):
    """Handle ChangeContext intents."""
    intent_type = INTENT_CHANGE_CONTEXT

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        if len(GROUP_ENTITIES) == 0:
            get_groups(hass)
        text = intent_obj.text_input.lower()
        _LOGGER.debug('text: ' + text)
        for idx, menu in enumerate(GROUP_ENTITIES, start=0):
            context_key_words = menu['context_key_words']
            if context_key_words is not None:
                context_key_words = context_key_words.split(',')
                if text in context_key_words:
                    set_curr_group(hass, menu)
                    set_curr_entity(hass, None)
                    message = menu['context_answer']
                    return message, True

        message = 'Nie znajduję odpowiedzi do kontekstu ' + text
        return message, False


class GetTimeIntent(intent.IntentHandler):
    """Handle GetTimeIntent intents."""
    intent_type = INTENT_GET_TIME

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        # hass = intent_obj.hass
        # time = hass.states.get('sensor.time').state
        # message = 'Jest godzina ' + time
        import babel.dates
        now = datetime.datetime.now()
        message = 'Jest ' + babel.dates.format_time(
            now, format='short', locale='pl')
        return message, True


class AisGetWeather(intent.IntentHandler):
    """Handle GetWeather intents."""
    intent_type = INTENT_GET_WEATHER

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        # weather = hass.states.get('sensor.pogoda_info').state
        weather = "Pogoda "
        attr = hass.states.get('group.ais_pogoda').attributes
        for a in attr['entity_id']:
            w = hass.states.get(a)
            if a == 'sensor.dark_sky_hourly_summary':
                weather += " " + w.state + " "
            elif a == 'sensor.dark_sky_daily_summary':
                weather += " " + w.state + " "
            else:
                weather += w.attributes['friendly_name'] + " " + w.state + " "
                if 'unit_of_measurement' in w.attributes:
                    if w.attributes['unit_of_measurement'] == 'hPa':
                        weather += "hektopascala; "
                    elif w.attributes['unit_of_measurement'] == 'km/h':
                        weather += "kilometrów na godzinę; "
                    elif w.attributes['unit_of_measurement'] == 'km':
                        weather += "kilometra; "
                    else:
                        weather += w.attributes['unit_of_measurement'] + "; "
        return weather, True


class AisGetWeather48(intent.IntentHandler):
    """Handle GetWeather48 intents."""
    intent_type = INTENT_GET_WEATHER_48

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        weather = hass.states.get('sensor.dark_sky_daily_summary').state
        return weather, True


class AisLampsOn(intent.IntentHandler):
    """Handle AisLampsOn intents."""
    intent_type = INTENT_LAMPS_ON

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
             'light', 'turn_on', {"entity_id": "group.all_lights"})
        return 'ok', True


class AisLampsOff(intent.IntentHandler):
    """Handle AisLampsOff intents."""
    intent_type = INTENT_LAMPS_OFF

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
             'light', 'turn_off', {"entity_id": "group.all_lights"})
        return 'ok', True


class AisSwitchesOn(intent.IntentHandler):
    """Handle AisSwitchesOn intents."""
    intent_type = INTENT_SWITCHES_ON

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
             'switch', 'turn_on', {"entity_id": "group.all_switches"})
        return 'ok', True


class AisSwitchesOff(intent.IntentHandler):
    """Handle AisSwitchesOff intents."""
    intent_type = INTENT_SWITCHES_OFF

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
             'switch', 'turn_off', {"entity_id": "group.all_switches"})
        return 'ok', True


class GetDateIntent(intent.IntentHandler):
    """Handle GetDateIntent intents."""
    intent_type = INTENT_GET_DATE

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        import babel.dates
        now = datetime.datetime.now()
        message = 'Jest ' + babel.dates.format_date(
            now, format='full', locale='pl')
        return message, True


class AisOpenCover(intent.IntentHandler):
    """Handle AisOpenCover intents."""
    intent_type = INTENT_OPEN_COVER
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        name = slots['item']['value']
        entity = _match_entity(hass, name)
        success = False

        if not entity:
            message = 'Nie znajduję urządzenia do otwarcia, o nazwie: ' + name
        else:
            # check if we can open on this device
            if entity.entity_id.startswith('cover.'):
                if entity.state == 'on':
                    # check if the device is already on
                    message = 'Urządzenie ' + name + ' jest już otwarte'
                elif entity.state == 'unavailable':
                    message = 'Urządzenie ' + name + ' jest niedostępne'
                else:
                    yield from hass.services.async_call(
                        core.DOMAIN, SERVICE_OPEN_COVER, {
                            ATTR_ENTITY_ID: entity.entity_id,
                        }, blocking=True)
                    message = 'OK, włączono {}'.format(entity.name)
                success = True
            else:
                message = 'Urządzenia ' + name + ' nie można otworzyć'
        return message, success


class AisCloseCover(intent.IntentHandler):
    """Handle AisCloseCover intents."""
    intent_type = INTENT_CLOSE_COVER
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle turn off intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        name = slots['item']['value']
        entity = _match_entity(hass, name)
        success = False
        if not entity:
            msg = 'Nie znajduję urządzenia do zamknięcia, o nazwie: ' + name
        else:
            # check if we can close on this device
            if entity.entity_id.startswith('cover.'):
                # check if the device is already closed
                if entity.state == 'off':
                    msg = 'Urządzenie {} jest już zamknięte'.format(
                        entity.name)
                elif entity.state == 'unavailable':
                    msg = 'Urządzenie {} jest niedostępne'.format(
                        entity.name)
                else:
                    yield from hass.services.async_call(
                        core.DOMAIN, SERVICE_TURN_OFF, {
                            ATTR_ENTITY_ID: entity.entity_id,
                        }, blocking=True)
                    msg = 'OK, zamknięto {}'.format(entity.name)
                    success = True
            else:
                msg = 'Urządzenia ' + name + ' nie można zamknąć'
        return msg, success


class AisStop(intent.IntentHandler):
    """Handle AisStop intents."""
    intent_type = INTENT_STOP

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'media_player', 'media_stop')
        message = 'ok, stop'
        return message, True


class AisPlay(intent.IntentHandler):
    """Handle AisPlay intents."""
    intent_type = INTENT_PLAY

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'media_player', 'media_play', {
                ATTR_ENTITY_ID: "media_player.wbudowany_glosnik",
            })
        message = 'ok, gram'
        return message, True


class AisNext(intent.IntentHandler):
    """Handle AisNext intents."""
    intent_type = INTENT_NEXT

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'media_player', 'media_next_track')
        message = 'ok, następny'
        return message, True


class AisPrev(intent.IntentHandler):
    """Handle AisPrev intents."""
    intent_type = INTENT_PREV

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'media_player', 'media_previous_track')
        message = 'ok, poprzedni'
        return message, True


class AisSceneActive(intent.IntentHandler):
    """Handle AisSceneActive intents."""
    intent_type = INTENT_SCENE
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        slots = self.async_validate_slots(intent_obj.slots)
        name = slots['item']['value']
        entity = _match_entity(hass, name)
        success = False

        if not entity:
            message = 'Nie znajduję sceny, o nazwie: ' + name
        else:
            # check if we can open on this device
            if entity.entity_id.startswith('scene.'):
                yield from hass.services.async_call(
                    'scene', 'turn_on', {
                        ATTR_ENTITY_ID: entity.entity_id,
                    }, blocking=True)
                message = 'OK, aktywuję {}'.format(entity.name)
                success = True
            else:
                message = name + ' nie można aktywować'
        return message, success


class AisSayIt(intent.IntentHandler):
    """Handle AisSayIt intents."""
    intent_type = INTENT_SAY_IT
    slot_schema = {
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        try:
            slots = self.async_validate_slots(intent_obj.slots)
            text = slots['item']['value']
        except Exception:
            text = None
        success = False
        if not text:
            import random
            answers = ['Nie wiem co mam powiedzieć?', 'Ale co?', 'Mówie mówie', 'OK, dobra zaraz coś wymyślę...',
                       'Mowa jest tylko srebrem', 'To samo czy coś nowego?']
            message = random.choice(answers)
        else:
            # check if we can open on this device
            message = text
            success = True
        return message, success


class AisClimateSetTemperature(intent.IntentHandler):
    """Handle AisClimateSetTemperature intents."""
    intent_type = INTENT_CLIMATE_SET_TEMPERATURE
    slot_schema = {
        'temp': cv.positive_int,
        'item': cv.string,
    }

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        try:
            hass = intent_obj.hass
            slots = self.async_validate_slots(intent_obj.slots)
            temp = slots['temp']['value']
            name = slots['item']['value']
            entity = _match_entity(hass, name)
        except Exception:
            text = None
        success = False
        if not entity:
            msg = 'Nie znajduję grzejnika, o nazwie: ' + name
        else:
            # check if we can close on this device
            if entity.entity_id.startswith('climate.'):
                # check if the device has already this temperature
                attr = hass.states.get(entity.entity_id).attributes
                if attr.get('temperature') == temp:
                    msg = '{} ma już ustawioną temperaturę {}'.format(
                        entity.name, temp)
                else:
                    yield from hass.services.async_call(
                        'climate', 'set_temperature', {
                            ATTR_ENTITY_ID: entity.entity_id,
                            "temperature": temp,
                            "target_temp_high": temp + 2,
                            "target_temp_low": temp - 6,
                            "operation_mode": "Heat"
                        }, blocking=True)
                    msg = 'OK, ustawiono temperaturę {} w {}'.format(temp, entity.name)
                    success = True
            else:
                msg = 'Na urządzeniu ' + name + ' nie można zmieniać temperatury.'
        return msg, success


class AisClimateSetAway(intent.IntentHandler):
    """Handle AisClimateSetAway intents."""
    intent_type = INTENT_CLIMATE_SET_AWAY

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'climate', 'set_away_mode', {"entity_id": "all", "away_mode": True})
        message = 'ok, tryb poza domem włączony'
        return message, True


class AisClimateUnSetAway(intent.IntentHandler):
    """Handle AisClimateUnSetAway intents."""
    intent_type = INTENT_CLIMATE_UNSET_AWAY

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'climate', 'set_away_mode', {"entity_id": "all", "away_mode": False})
        message = 'ok, tryb poza domem wyłączony'
        return message, True


class AisClimateSetAllOn(intent.IntentHandler):
    """Handle AisClimateSetAllOn intents."""
    intent_type = INTENT_CLIMATE_SET_ALL_ON

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'climate', 'set_operation_mode', {"entity_id": "all", 'operation_mode': 'heat'})
        message = 'ok, całe ogrzewanie włączone'
        return message, True


class AisClimateSetAllOff(intent.IntentHandler):
    """Handle AisClimateSetAllOff intents."""
    intent_type = INTENT_CLIMATE_SET_ALL_OFF

    @asyncio.coroutine
    def async_handle(self, intent_obj):
        """Handle the intent."""
        hass = intent_obj.hass
        yield from hass.services.async_call(
            'climate', 'set_operation_mode', {"entity_id": "all", 'operation_mode': 'off'})
        message = 'ok, całe ogrzewanie wyłączone'
        return message, True

# class AisClimateSetOff(intent.IntentHandler):
#     """Handle AisClimateSetOff intents."""
#     intent_type = INTENT_CLIMATE_SET_OFF
#     slot_schema = {
#         'item': cv.string,
#     }
#
#     @asyncio.coroutine
#     def async_handle(self, intent_obj):
#         """Handle the intent."""
#         try:
#             hass = intent_obj.hass
#             slots = self.async_validate_slots(intent_obj.slots)
#             name = slots['item']['value']
#             entity = _match_entity(hass, name)
#         except Exception:
#             text = None
#         success = False
#         if not entity:
#             msg = 'Nie znajduję grzejnika, o nazwie: ' + name
#         else:
#             # check if we can close on this device
#             if entity.entity_id.startswith('climate.'):
#             # check if the device already is off
#                 if entity.state == 'off':
#                     msg = '{} jest już wyłączony'.format(entity.name)
#                 elif entity.state == 'unavailable':
#                     msg = '{} jest niedostępny'.format(entity.name)
#                 else:
#                     yield from hass.services.async_call(
#                         'climate', 'turn_off', {
#                             ATTR_ENTITY_ID: entity.entity_id
#                         }, blocking=True)
#                     msg = 'OK, wyłączono ogrzewanie {} '.format(entity.name)
#                     success = True
#             else:
#                 msg = 'Na urządzeniu ' + name + ' nie można wyłączyć ogrzwania.'
#             return msg, success
#
#
# class AisClimateSetOn(intent.IntentHandler):
#     """Handle AisClimateSetOn intents."""
#     intent_type = INTENT_CLIMATE_SET_ON
#     slot_schema = {
#         'item': cv.string,
#     }
#
#     @asyncio.coroutine
#     def async_handle(self, intent_obj):
#         """Handle the intent."""
#         try:
#             hass = intent_obj.hass
#             slots = self.async_validate_slots(intent_obj.slots)
#             name = slots['item']['value']
#             entity = _match_entity(hass, name)
#         except Exception:
#             text = None
#         success = False
#         if not entity:
#             msg = 'Nie znajduję grzejnika, o nazwie: ' + name
#         else:
#             # check if we can close on this device
#             if entity.entity_id.startswith('climate.'):
#                 # check if the device already is heat
#                 if entity.state == 'heat':
#                     msg = '{} jest już włączony'.format(entity.name)
#                 elif entity.state == 'unavailable':
#                     msg = '{} jest niedostępny'.format(entity.name)
#                 else:
#                     yield from hass.services.async_call(
#                         'climate', 'turn_on', {
#                             ATTR_ENTITY_ID: entity.entity_id
#                         }, blocking=True)
#                     msg = 'OK, wyłączono ogrzewanie {} '.format(entity.name)
#                     success = True
#             else:
#                 msg = 'Na urządzeniu ' + name + ' nie można wyłączyć ogrzwania.'
#             return msg, success



"""
Support for Google Home bluetooth tacker.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/device_tracker.googlehome/
"""
import logging

import voluptuous as vol

import homeassistant.helpers.config_validation as cv
from homeassistant.helpers.aiohttp_client import async_get_clientsession
from homeassistant.components.device_tracker import (
    DOMAIN, PLATFORM_SCHEMA, DeviceScanner)
from homeassistant.const import CONF_HOST

REQUIREMENTS = ['ghlocalapi==0.3.5']

_LOGGER = logging.getLogger(__name__)

CONF_RSSI_THRESHOLD = 'rssi_threshold'

PLATFORM_SCHEMA = vol.All(
    PLATFORM_SCHEMA.extend({
        vol.Required(CONF_HOST): cv.string,
        vol.Optional(CONF_RSSI_THRESHOLD, default=-70): vol.Coerce(int),
    }))


async def async_get_scanner(hass, config):
    """Validate the configuration and return an Google Home scanner."""
    scanner = GoogleHomeDeviceScanner(hass, config[DOMAIN])
    await scanner.async_connect()
    return scanner if scanner.success_init else None


class GoogleHomeDeviceScanner(DeviceScanner):
    """This class queries a Google Home unit."""

    def __init__(self, hass, config):
        """Initialize the scanner."""
        from ghlocalapi.device_info import DeviceInfo
        from ghlocalapi.bluetooth import Bluetooth

        self.last_results = {}

        self.success_init = False
        self._host = config[CONF_HOST]
        self.rssi_threshold = config[CONF_RSSI_THRESHOLD]

        session = async_get_clientsession(hass)
        self.deviceinfo = DeviceInfo(hass.loop, session, self._host)
        self.scanner = Bluetooth(hass.loop, session, self._host)

    async def async_connect(self):
        """Initialize connection to Google Home."""
        await self.deviceinfo.get_device_info()
        data = self.deviceinfo.device_info
        self.success_init = data is not None

    async def async_scan_devices(self):
        """Scan for new devices and return a list with found device IDs."""
        await self.async_update_info()
        return list(self.last_results.keys())

    async def async_get_device_name(self, device):
        """Return the name of the given device or None if we don't know."""
        if device not in self.last_results:
            return None
        return '{}_{}'.format(self._host,
                              self.last_results[device]['btle_mac_address'])

    async def get_extra_attributes(self, device):
        """Return the extra attributes of the device."""
        return self.last_results[device]

    async def async_update_info(self):
        """Ensure the information from Google Home is up to date."""
        _LOGGER.debug('Checking Devices...')
        await self.scanner.scan_for_devices()
        await self.scanner.get_scan_result()
        ghname = self.deviceinfo.device_info['name']
        devices = {}
        for device in self.scanner.devices:
            if device['rssi'] > self.rssi_threshold:
                uuid = '{}_{}'.format(self._host, device['mac_address'])
                devices[uuid] = {}
                devices[uuid]['rssi'] = device['rssi']
                devices[uuid]['btle_mac_address'] = device['mac_address']
                devices[uuid]['ghname'] = ghname
                devices[uuid]['source_type'] = 'bluetooth'
        await self.scanner.clear_scan_result()
        self.last_results = devices



"""
This component provides support to the Ring Door Bell camera.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/camera.ring/
"""
import asyncio
import logging

from datetime import timedelta

import voluptuous as vol

from homeassistant.helpers import config_validation as cv
from homeassistant.components.ring import (
    DATA_RING, CONF_ATTRIBUTION, NOTIFICATION_ID)
from homeassistant.components.camera import Camera, PLATFORM_SCHEMA
from homeassistant.components.ffmpeg import DATA_FFMPEG
from homeassistant.const import ATTR_ATTRIBUTION, CONF_SCAN_INTERVAL
from homeassistant.helpers.aiohttp_client import async_aiohttp_proxy_stream
from homeassistant.util import dt as dt_util

CONF_FFMPEG_ARGUMENTS = 'ffmpeg_arguments'

DEPENDENCIES = ['ring', 'ffmpeg']

FORCE_REFRESH_INTERVAL = timedelta(minutes=45)

_LOGGER = logging.getLogger(__name__)

NOTIFICATION_TITLE = 'Ring Camera Setup'

SCAN_INTERVAL = timedelta(seconds=90)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Optional(CONF_FFMPEG_ARGUMENTS): cv.string,
    vol.Optional(CONF_SCAN_INTERVAL, default=SCAN_INTERVAL):
        cv.time_period,
})


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up a Ring Door Bell and StickUp Camera."""
    ring = hass.data[DATA_RING]

    cams = []
    cams_no_plan = []
    for camera in ring.doorbells:
        if camera.has_subscription:
            cams.append(RingCam(hass, camera, config))
        else:
            cams_no_plan.append(camera)

    for camera in ring.stickup_cams:
        if camera.has_subscription:
            cams.append(RingCam(hass, camera, config))
        else:
            cams_no_plan.append(camera)

    # show notification for all cameras without an active subscription
    if cams_no_plan:
        cameras = str(', '.join([camera.name for camera in cams_no_plan]))

        err_msg = '''A Ring Protect Plan is required for the''' \
                  ''' following cameras: {}.'''.format(cameras)

        _LOGGER.error(err_msg)
        hass.components.persistent_notification.create(
            'Error: {}<br />'
            'You will need to restart hass after fixing.'
            ''.format(err_msg),
            title=NOTIFICATION_TITLE,
            notification_id=NOTIFICATION_ID)

    add_entities(cams, True)
    return True


class RingCam(Camera):
    """An implementation of a Ring Door Bell camera."""

    def __init__(self, hass, camera, device_info):
        """Initialize a Ring Door Bell camera."""
        super(RingCam, self).__init__()
        self._camera = camera
        self._hass = hass
        self._name = self._camera.name
        self._ffmpeg = hass.data[DATA_FFMPEG]
        self._ffmpeg_arguments = device_info.get(CONF_FFMPEG_ARGUMENTS)
        self._last_video_id = self._camera.last_recording_id
        self._video_url = self._camera.recording_url(self._last_video_id)
        self._utcnow = dt_util.utcnow()
        self._expires_at = FORCE_REFRESH_INTERVAL + self._utcnow

    @property
    def name(self):
        """Return the name of this camera."""
        return self._name

    @property
    def unique_id(self):
        """Return a unique ID."""
        return self._camera.id

    @property
    def device_state_attributes(self):
        """Return the state attributes."""
        return {
            ATTR_ATTRIBUTION: CONF_ATTRIBUTION,
            'device_id': self._camera.id,
            'firmware': self._camera.firmware,
            'kind': self._camera.kind,
            'timezone': self._camera.timezone,
            'type': self._camera.family,
            'video_url': self._video_url,
        }

    async def async_camera_image(self):
        """Return a still image response from the camera."""
        from haffmpeg import ImageFrame, IMAGE_JPEG
        ffmpeg = ImageFrame(self._ffmpeg.binary, loop=self.hass.loop)

        if self._video_url is None:
            return

        image = await asyncio.shield(ffmpeg.get_image(
            self._video_url, output_format=IMAGE_JPEG,
            extra_cmd=self._ffmpeg_arguments), loop=self.hass.loop)
        return image

    async def handle_async_mjpeg_stream(self, request):
        """Generate an HTTP MJPEG stream from the camera."""
        from haffmpeg import CameraMjpeg

        if self._video_url is None:
            return

        stream = CameraMjpeg(self._ffmpeg.binary, loop=self.hass.loop)
        await stream.open_camera(
            self._video_url, extra_cmd=self._ffmpeg_arguments)

        try:
            return await async_aiohttp_proxy_stream(
                self.hass, request, stream,
                'multipart/x-mixed-replace;boundary=ffserver')
        finally:
            await stream.close()

    @property
    def should_poll(self):
        """Update the image periodically."""
        return True

    def update(self):
        """Update camera entity and refresh attributes."""
        _LOGGER.debug("Checking if Ring DoorBell needs to refresh video_url")

        self._camera.update()
        self._utcnow = dt_util.utcnow()

        last_recording_id = self._camera.last_recording_id

        if self._last_video_id != last_recording_id or \
           self._utcnow >= self._expires_at:

            _LOGGER.info("Ring DoorBell properties refreshed")

            # update attributes if new video or if URL has expired
            self._last_video_id = self._camera.last_recording_id
            self._video_url = self._camera.recording_url(self._last_video_id)
            self._expires_at = FORCE_REFRESH_INTERVAL + self._utcnow



"""Define constants for the SimpliSafe component."""
import logging
from datetime import timedelta

LOGGER = logging.getLogger('homeassistant.components.rainmachine')

DOMAIN = 'rainmachine'

DATA_CLIENT = 'client'

DEFAULT_PORT = 8080
DEFAULT_SCAN_INTERVAL = timedelta(seconds=60)
DEFAULT_SSL = True

TOPIC_UPDATE = 'update_{0}'



"""
Support for Tellstick switches using Tellstick Net.

This platform uses the Telldus Live online service.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/light.tellduslive/
"""
import logging

from homeassistant.components import light, tellduslive
from homeassistant.components.light import (
    ATTR_BRIGHTNESS, SUPPORT_BRIGHTNESS, Light)
from homeassistant.components.tellduslive.entry import TelldusLiveEntity
from homeassistant.helpers.dispatcher import async_dispatcher_connect

_LOGGER = logging.getLogger(__name__)


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Old way of setting up TelldusLive.

    Can only be called when a user accidentally mentions the platform in their
    config. But even in that case it would have been ignored.
    """
    pass


async def async_setup_entry(hass, config_entry, async_add_entities):
    """Set up tellduslive sensors dynamically."""
    async def async_discover_light(device_id):
        """Discover and add a discovered sensor."""
        client = hass.data[tellduslive.DOMAIN]
        async_add_entities([TelldusLiveLight(client, device_id)])

    async_dispatcher_connect(
        hass,
        tellduslive.TELLDUS_DISCOVERY_NEW.format(light.DOMAIN,
                                                 tellduslive.DOMAIN),
        async_discover_light,
    )


class TelldusLiveLight(TelldusLiveEntity, Light):
    """Representation of a Tellstick Net light."""

    def __init__(self, client, device_id):
        """Initialize the  Tellstick Net light."""
        super().__init__(client, device_id)
        self._last_brightness = self.brightness

    def changed(self):
        """Define a property of the device that might have changed."""
        self._last_brightness = self.brightness

    @property
    def brightness(self):
        """Return the brightness of this light between 0..255."""
        return self.device.dim_level

    @property
    def supported_features(self):
        """Flag supported features."""
        return SUPPORT_BRIGHTNESS

    @property
    def is_on(self):
        """Return true if light is on."""
        return self.device.is_on

    def turn_on(self, **kwargs):
        """Turn the light on."""
        brightness = kwargs.get(ATTR_BRIGHTNESS, self._last_brightness)
        self.device.dim(level=brightness)
        self.changed()

    def turn_off(self, **kwargs):
        """Turn the light off."""
        self.device.turn_off()
        self.changed()



"""
Support for Songpal-enabled (Sony) media devices.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/media_player.songpal/
"""
import asyncio
import logging
from collections import OrderedDict

import voluptuous as vol

from homeassistant.components.media_player import (
    DOMAIN, PLATFORM_SCHEMA, SUPPORT_SELECT_SOURCE, SUPPORT_TURN_OFF,
    SUPPORT_TURN_ON, SUPPORT_VOLUME_MUTE, SUPPORT_VOLUME_SET,
    SUPPORT_VOLUME_STEP, MediaPlayerDevice)
from homeassistant.const import (
    ATTR_ENTITY_ID, CONF_NAME, STATE_OFF, STATE_ON, EVENT_HOMEASSISTANT_STOP)
from homeassistant.exceptions import PlatformNotReady
import homeassistant.helpers.config_validation as cv

REQUIREMENTS = ['python-songpal==0.0.9.1']

_LOGGER = logging.getLogger(__name__)

CONF_ENDPOINT = 'endpoint'

PARAM_NAME = 'name'
PARAM_VALUE = 'value'

PLATFORM = 'songpal'

SET_SOUND_SETTING = 'songpal_set_sound_setting'

SUPPORT_SONGPAL = SUPPORT_VOLUME_SET | SUPPORT_VOLUME_STEP | \
                  SUPPORT_VOLUME_MUTE | SUPPORT_SELECT_SOURCE | \
                  SUPPORT_TURN_ON | SUPPORT_TURN_OFF

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Optional(CONF_NAME): cv.string,
    vol.Required(CONF_ENDPOINT): cv.string,
})

SET_SOUND_SCHEMA = vol.Schema({
    vol.Optional(ATTR_ENTITY_ID): cv.entity_id,
    vol.Required(PARAM_NAME): cv.string,
    vol.Required(PARAM_VALUE): cv.string,
})


async def async_setup_platform(
        hass, config, async_add_entities, discovery_info=None):
    """Set up the Songpal platform."""
    from songpal import SongpalException

    if PLATFORM not in hass.data:
        hass.data[PLATFORM] = {}

    if discovery_info is not None:
        name = discovery_info["name"]
        endpoint = discovery_info["properties"]["endpoint"]
        _LOGGER.debug("Got autodiscovered %s - endpoint: %s", name, endpoint)

        device = SongpalDevice(name, endpoint)
    else:
        name = config.get(CONF_NAME)
        endpoint = config.get(CONF_ENDPOINT)
        device = SongpalDevice(name, endpoint, poll=False)

    if endpoint in hass.data[PLATFORM]:
        _LOGGER.debug("The endpoint exists already, skipping setup.")
        return

    try:
        await device.initialize()
    except SongpalException as ex:
        _LOGGER.error("Unable to get methods from songpal: %s", ex)
        raise PlatformNotReady

    hass.data[PLATFORM][endpoint] = device

    async_add_entities([device], True)

    async def async_service_handler(service):
        """Service handler."""
        entity_id = service.data.get("entity_id", None)
        params = {key: value for key, value in service.data.items()
                  if key != ATTR_ENTITY_ID}

        for device in hass.data[PLATFORM].values():
            if device.entity_id == entity_id or entity_id is None:
                _LOGGER.debug("Calling %s (entity: %s) with params %s",
                              service, entity_id, params)

                await device.async_set_sound_setting(
                    params[PARAM_NAME], params[PARAM_VALUE])

    hass.services.async_register(
        DOMAIN, SET_SOUND_SETTING, async_service_handler,
        schema=SET_SOUND_SCHEMA)


class SongpalDevice(MediaPlayerDevice):
    """Class representing a Songpal device."""

    def __init__(self, name, endpoint, poll=False):
        """Init."""
        from songpal import Device
        self._name = name
        self._endpoint = endpoint
        self._poll = poll
        self.dev = Device(self._endpoint)
        self._sysinfo = None

        self._state = False
        self._available = False
        self._initialized = False

        self._volume_control = None
        self._volume_min = 0
        self._volume_max = 1
        self._volume = 0
        self._is_muted = False

        self._active_source = None
        self._sources = {}

    @property
    def should_poll(self):
        """Return True if the device should be polled."""
        return self._poll

    async def initialize(self):
        """Initialize the device."""
        await self.dev.get_supported_methods()
        self._sysinfo = await self.dev.get_system_info()

    async def async_activate_websocket(self):
        """Activate websocket for listening if wanted."""
        _LOGGER.info("Activating websocket connection..")
        from songpal import (VolumeChange, ContentChange,
                             PowerChange, ConnectChange)

        async def _volume_changed(volume: VolumeChange):
            _LOGGER.debug("Volume changed: %s", volume)
            self._volume = volume.volume
            self._is_muted = volume.mute
            await self.async_update_ha_state()

        async def _source_changed(content: ContentChange):
            _LOGGER.debug("Source changed: %s", content)
            if content.is_input:
                self._active_source = self._sources[content.source]
                _LOGGER.debug("New active source: %s", self._active_source)
                await self.async_update_ha_state()
            else:
                _LOGGER.warning("Got non-handled content change: %s",
                                content)

        async def _power_changed(power: PowerChange):
            _LOGGER.debug("Power changed: %s", power)
            self._state = power.status
            await self.async_update_ha_state()

        async def _try_reconnect(connect: ConnectChange):
            _LOGGER.error("Got disconnected with %s, trying to reconnect.",
                          connect.exception)
            self._available = False
            self.dev.clear_notification_callbacks()
            await self.async_update_ha_state()

            # Try to reconnect forever, a successful reconnect will initialize
            # the websocket connection again.
            delay = 10
            while not self._available:
                _LOGGER.debug("Trying to reconnect in %s seconds", delay)
                await asyncio.sleep(delay)
                # We need to inform HA about the state in case we are coming
                # back from a disconnected state.
                await self.async_update_ha_state(force_refresh=True)
                delay = min(2*delay, 300)

        self.dev.on_notification(VolumeChange, _volume_changed)
        self.dev.on_notification(ContentChange, _source_changed)
        self.dev.on_notification(PowerChange, _power_changed)
        self.dev.on_notification(ConnectChange, _try_reconnect)

        async def listen_events():
            await self.dev.listen_notifications()

        async def handle_stop(event):
            await self.dev.stop_listen_notifications()

        self.hass.bus.async_listen_once(EVENT_HOMEASSISTANT_STOP, handle_stop)

        self.hass.loop.create_task(listen_events())

    @property
    def name(self):
        """Return name of the device."""
        return self._name

    @property
    def unique_id(self):
        """Return a unique ID."""
        return self._sysinfo.macAddr

    @property
    def available(self):
        """Return availability of the device."""
        return self._available

    async def async_set_sound_setting(self, name, value):
        """Change a setting on the device."""
        await self.dev.set_sound_settings(name, value)

    async def async_update(self):
        """Fetch updates from the device."""
        from songpal import SongpalException
        try:
            volumes = await self.dev.get_volume_information()
            if not volumes:
                _LOGGER.error("Got no volume controls, bailing out")
                self._available = False
                return

            if len(volumes) > 1:
                _LOGGER.debug(
                    "Got %s volume controls, using the first one", volumes)

            volume = volumes[0]
            _LOGGER.debug("Current volume: %s", volume)

            self._volume_max = volume.maxVolume
            self._volume_min = volume.minVolume
            self._volume = volume.volume
            self._volume_control = volume
            self._is_muted = self._volume_control.is_muted

            status = await self.dev.get_power()
            self._state = status.status
            _LOGGER.debug("Got state: %s", status)

            inputs = await self.dev.get_inputs()
            _LOGGER.debug("Got ins: %s", inputs)

            self._sources = OrderedDict()
            for input_ in inputs:
                self._sources[input_.uri] = input_
                if input_.active:
                    self._active_source = input_

            _LOGGER.debug("Active source: %s", self._active_source)

            self._available = True

            # activate notifications if wanted
            if not self._poll:
                await self.hass.async_create_task(
                    self.async_activate_websocket())
        except SongpalException as ex:
            _LOGGER.error("Unable to update: %s", ex)
            self._available = False

    async def async_select_source(self, source):
        """Select source."""
        for out in self._sources.values():
            if out.title == source:
                await out.activate()
                return

        _LOGGER.error("Unable to find output: %s", source)

    @property
    def source_list(self):
        """Return list of available sources."""
        return [src.title for src in self._sources.values()]

    @property
    def state(self):
        """Return current state."""
        if self._state:
            return STATE_ON
        return STATE_OFF

    @property
    def source(self):
        """Return currently active source."""
        return self._active_source.title

    @property
    def volume_level(self):
        """Return volume level."""
        volume = self._volume / self._volume_max
        return volume

    async def async_set_volume_level(self, volume):
        """Set volume level."""
        volume = int(volume * self._volume_max)
        _LOGGER.debug("Setting volume to %s", volume)
        return await self._volume_control.set_volume(volume)

    async def async_volume_up(self):
        """Set volume up."""
        return await self._volume_control.set_volume("+1")

    async def async_volume_down(self):
        """Set volume down."""
        return await self._volume_control.set_volume("-1")

    async def async_turn_on(self):
        """Turn the device on."""
        return await self.dev.set_power(True)

    async def async_turn_off(self):
        """Turn the device off."""
        return await self.dev.set_power(False)

    async def async_mute_volume(self, mute):
        """Mute or unmute the device."""
        _LOGGER.debug("Set mute: %s", mute)
        return await self._volume_control.set_mute(mute)

    @property
    def is_volume_muted(self):
        """Return whether the device is muted."""
        return self._is_muted

    @property
    def supported_features(self):
        """Return supported features."""
        return SUPPORT_SONGPAL



"""
Sensor to get GTT's timetable for a stop.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/sensor.gtt/
"""
import logging
from datetime import timedelta, datetime

import voluptuous as vol

from homeassistant.components.sensor import PLATFORM_SCHEMA
from homeassistant.const import DEVICE_CLASS_TIMESTAMP
from homeassistant.helpers.entity import Entity
import homeassistant.helpers.config_validation as cv
import homeassistant.util.dt as dt_util

REQUIREMENTS = ['pygtt==1.1.2']

_LOGGER = logging.getLogger(__name__)

CONF_STOP = 'stop'
CONF_BUS_NAME = 'bus_name'

ICON = 'mdi:train'

SCAN_INTERVAL = timedelta(minutes=2)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Required(CONF_STOP): cv.string,
    vol.Optional(CONF_BUS_NAME): cv.string
})


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up the Gtt platform."""
    stop = config[CONF_STOP]
    bus_name = config.get(CONF_BUS_NAME)

    add_entities([GttSensor(stop, bus_name)], True)


class GttSensor(Entity):
    """Representation of a Gtt Sensor."""

    def __init__(self, stop, bus_name):
        """Initialize the Gtt sensor."""
        self.data = GttData(stop, bus_name)
        self._state = None
        self._name = 'Stop {}'.format(stop)

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name

    @property
    def icon(self):
        """Return the icon of the sensor."""
        return ICON

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def device_class(self):
        """Return the device class."""
        return DEVICE_CLASS_TIMESTAMP

    @property
    def device_state_attributes(self):
        """Return the state attributes of the sensor."""
        attr = {
            'bus_name': self.data.state_bus['bus_name']
        }
        return attr

    def update(self):
        """Update device state."""
        self.data.get_data()
        next_time = dt_util.parse_time(self.data.state_bus['time'][0]['run'])
        self._state = next_time.isoformat()


class GttData:
    """Inteface to PyGTT."""

    def __init__(self, stop, bus_name):
        """Initialize the GttData class."""
        from pygtt import PyGTT
        self._pygtt = PyGTT()
        self._stop = stop
        self._bus_name = bus_name
        self.bus_list = {}
        self.state_bus = {}

    def get_data(self):
        """Get the data from the api."""
        self.bus_list = self._pygtt.get_by_stop(self._stop)
        self.bus_list.sort(key=lambda b:
                           datetime.strptime(b['time'][0]['run'], "%H:%M"))

        if self._bus_name is not None:
            self.state_bus = self.get_bus_by_name()
            return

        self.state_bus = self.bus_list[0]

    def get_bus_by_name(self):
        """Get the bus by name."""
        for bus in self.bus_list:
            if bus['bus_name'] == self._bus_name:
                return bus



"""
Support for Ambient Weather Station Service.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/sensor.ambient_station/
"""

import asyncio
from datetime import timedelta
import logging

import voluptuous as vol

from homeassistant.components.sensor import PLATFORM_SCHEMA
from homeassistant.const import CONF_API_KEY, CONF_MONITORED_CONDITIONS
import homeassistant.helpers.config_validation as cv
from homeassistant.helpers.entity import Entity
from homeassistant.util import Throttle

REQUIREMENTS = ['ambient_api==1.5.2']

CONF_APP_KEY = 'app_key'

SENSOR_NAME = 0
SENSOR_UNITS = 1

CONF_UNITS = 'units'
UNITS_US = 'us'
UNITS_SI = 'si'
UNIT_SYSTEM = {UNITS_US: 0, UNITS_SI: 1}

SCAN_INTERVAL = timedelta(seconds=300)

SENSOR_TYPES = {
    'winddir': ['Wind Dir', 'º'],
    'windspeedmph': ['Wind Speed', 'mph'],
    'windgustmph': ['Wind Gust', 'mph'],
    'maxdailygust': ['Max Gust', 'mph'],
    'windgustdir': ['Gust Dir', 'º'],
    'windspdmph_avg2m': ['Wind Avg 2m', 'mph'],
    'winddir_avg2m': ['Wind Dir Avg 2m', 'mph'],
    'windspdmph_avg10m': ['Wind Avg 10m', 'mph'],
    'winddir_avg10m': ['Wind Dir Avg 10m', 'º'],
    'humidity': ['Humidity', '%'],
    'humidityin': ['Humidity In', '%'],
    'tempf': ['Temp', ['ºF', 'ºC']],
    'tempinf': ['Inside Temp', ['ºF', 'ºC']],
    'battout': ['Battery', ''],
    'hourlyrainin': ['Hourly Rain Rate', 'in/hr'],
    'dailyrainin': ['Daily Rain', 'in'],
    '24hourrainin': ['24 Hr Rain', 'in'],
    'weeklyrainin': ['Weekly Rain', 'in'],
    'monthlyrainin': ['Monthly Rain', 'in'],
    'yearlyrainin': ['Yearly Rain', 'in'],
    'eventrainin': ['Event Rain', 'in'],
    'totalrainin': ['Lifetime Rain', 'in'],
    'baromrelin': ['Rel Pressure', 'inHg'],
    'baromabsin': ['Abs Pressure', 'inHg'],
    'uv': ['uv', 'Index'],
    'solarradiation': ['Solar Rad', 'W/m^2'],
    'co2': ['co2', 'ppm'],
    'lastRain': ['Last Rain', ''],
    'dewPoint': ['Dew Point', ['ºF', 'ºC']],
    'feelsLike': ['Feels Like', ['ºF', 'ºC']],
}

_LOGGER = logging.getLogger(__name__)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Required(CONF_API_KEY): cv.string,
    vol.Required(CONF_APP_KEY): cv.string,
    vol.Required(CONF_MONITORED_CONDITIONS):
        vol.All(cv.ensure_list, [vol.In(SENSOR_TYPES)]),
    vol.Optional(CONF_UNITS): vol.In([UNITS_SI, UNITS_US]),
})


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Initialze each sensor platform for each monitored condition."""
    api_key = config[CONF_API_KEY]
    app_key = config[CONF_APP_KEY]
    station_data = AmbientStationData(hass, api_key, app_key)
    if not station_data.connect_success:
        _LOGGER.error("Could not connect to weather station API")
        return

    sensor_list = []

    if CONF_UNITS in config:
        sys_units = config[CONF_UNITS]
    elif hass.config.units.is_metric:
        sys_units = UNITS_SI
    else:
        sys_units = UNITS_US

    for condition in config[CONF_MONITORED_CONDITIONS]:
        # create a sensor object for each monitored condition
        sensor_params = SENSOR_TYPES[condition]
        name = sensor_params[SENSOR_NAME]
        units = sensor_params[SENSOR_UNITS]
        if isinstance(units, list):
            units = sensor_params[SENSOR_UNITS][UNIT_SYSTEM[sys_units]]

        sensor_list.append(AmbientWeatherSensor(station_data, condition,
                                                name, units))

    add_entities(sensor_list)


class AmbientWeatherSensor(Entity):
    """Representation of a Sensor."""

    def __init__(self, station_data, condition, name, units):
        """Initialize the sensor."""
        self._state = None
        self.station_data = station_data
        self._condition = condition
        self._name = name
        self._units = units

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def unit_of_measurement(self):
        """Return the unit of measurement."""
        return self._units

    async def async_update(self):
        """Fetch new state data for the sensor.

        This is the only method that should fetch new data for Home Assistant.
        """
        _LOGGER.debug("Getting data for sensor: %s", self._name)
        data = await self.station_data.get_data()
        if data is None:
            # update likely got throttled and returned None, so use the cached
            # data from the station_data object
            self._state = self.station_data.data[self._condition]
        else:
            if self._condition in data:
                self._state = data[self._condition]
            else:
                _LOGGER.warning("%s sensor data not available from the "
                                "station", self._condition)

        _LOGGER.debug("Sensor: %s | Data: %s", self._name, self._state)


class AmbientStationData:
    """Class to interface with ambient-api library."""

    def __init__(self, hass, api_key, app_key):
        """Initialize station data object."""
        self.hass = hass
        self._api_keys = {
            'AMBIENT_ENDPOINT':
            'https://api.ambientweather.net/v1',
            'AMBIENT_API_KEY': api_key,
            'AMBIENT_APPLICATION_KEY': app_key,
            'log_level': 'DEBUG'
        }

        self.data = None
        self._station = None
        self._api = None
        self._devices = None
        self.connect_success = False

        self.get_data = Throttle(SCAN_INTERVAL)(self.async_update)
        self._connect_api()     # attempt to connect to API

    async def async_update(self):
        """Get new data."""
        # refresh API connection since servers turn over nightly
        _LOGGER.debug("Getting new data from server")
        new_data = None
        await self.hass.async_add_executor_job(self._connect_api)
        await asyncio.sleep(2)   # need minimum 2 seconds between API calls
        if self._station is not None:
            data = await self.hass.async_add_executor_job(
                self._station.get_data)
            if data is not None:
                new_data = data[0]
                self.data = new_data
            else:
                _LOGGER.debug("data is None type")
        else:
            _LOGGER.debug("Station is None type")

        return new_data

    def _connect_api(self):
        """Connect to the API and capture new data."""
        from ambient_api.ambientapi import AmbientAPI

        self._api = AmbientAPI(**self._api_keys)
        self._devices = self._api.get_devices()

        if self._devices:
            self._station = self._devices[0]
            if self._station is not None:
                self.connect_success = True
        else:
            _LOGGER.debug("No station devices available")



"""
A sensor platform that give you information about the next space launch.

For more details about this platform, please refer to the documentation at
https://www.home-assistant.io/components/sensor.launch_library/
"""
from datetime import timedelta
import logging

import voluptuous as vol

import homeassistant.helpers.config_validation as cv
from homeassistant.components.sensor import PLATFORM_SCHEMA
from homeassistant.const import ATTR_ATTRIBUTION, CONF_NAME
from homeassistant.helpers.entity import Entity
from homeassistant.helpers.aiohttp_client import async_get_clientsession

REQUIREMENTS = ['pylaunches==0.2.0']

_LOGGER = logging.getLogger(__name__)

ATTRIBUTION = "Data provided by Launch Library."

DEFAULT_NAME = 'Next launch'

SCAN_INTERVAL = timedelta(hours=1)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend({
    vol.Optional(CONF_NAME, default=DEFAULT_NAME): cv.string,
    })


async def async_setup_platform(
        hass, config, async_add_entities, discovery_info=None):
    """Create the launch sensor."""
    from pylaunches.api import Launches

    name = config[CONF_NAME]

    session = async_get_clientsession(hass)
    launches = Launches(hass.loop, session)
    sensor = [LaunchLibrarySensor(launches, name)]
    async_add_entities(sensor, True)


class LaunchLibrarySensor(Entity):
    """Representation of a launch_library Sensor."""

    def __init__(self, launches, name):
        """Initialize the sensor."""
        self.launches = launches
        self._attributes = {}
        self._name = name
        self._state = None

    async def async_update(self):
        """Get the latest data."""
        await self.launches.get_launches()
        if self.launches.launches is None:
            _LOGGER.error("No data recieved")
            return
        try:
            data = self.launches.launches[0]
            self._state = data['name']
            self._attributes['launch_time'] = data['start']
            self._attributes['agency'] = data['agency']
            agency_country_code = data['agency_country_code']
            self._attributes['agency_country_code'] = agency_country_code
            self._attributes['stream'] = data['stream']
            self._attributes[ATTR_ATTRIBUTION] = ATTRIBUTION
        except (KeyError, IndexError) as error:
            _LOGGER.debug("Error getting data, %s", error)

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def icon(self):
        """Return the icon of the sensor."""
        return 'mdi:rocket'

    @property
    def device_state_attributes(self):
        """Return attributes for the sensor."""
        return self._attributes
